"""
Cache of loop statuses.
"""

from __future__ import annotations
from dataclasses import dataclass
from collections import defaultdict
from typing import TYPE_CHECKING
from typing_extensions import Generic, TypeVar

from hpcflow.sdk.core.utils import nth_key
from hpcflow.sdk.log import TimeIt
from hpcflow.sdk.core.cache import ObjectCache

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from typing_extensions import Self
    from ..typing import DataIndex
    from .loop import Loop
    from .task import WorkflowTask
    from .types import DependentDescriptor, ElementDescriptor
    from .workflow import Workflow

K = TypeVar("K")
V = TypeVar("V")


class _LoopIndexError(TypeError):
    """
    A type error special to loop indices.
    """

    def __init__(self, loop_index: LoopIndex) -> None:
        super().__init__(
            f"{loop_index.__class__.__name__} does not support item assignment"
        )


class LoopIndex(dict[K, V], Generic[K, V]):
    """
    Hashable dict implementation, suitable for use as a key into
    other dicts. Once used as a key, becomes immutable.

    Example
    -------

        >>> h1 = LoopIndex({"apples": 1, "bananas":2})
        >>> h2 = LoopIndex({"bananas": 3, "mangoes": 5})
        >>> h1+h2
        LoopIndex(apples=1, bananas=3, mangoes=5)
        >>> d1 = {}
        >>> d1[h1] = "salad"
        >>> d1[h1]
        'salad'
        >>> d1[h2]
        Traceback (most recent call last):
        ...
        KeyError: LoopIndex(bananas=3, mangoes=5)

    Notes
    -----
    * Based on answers from
      http://stackoverflow.com/questions/1151658/python-hashable-dicts
    * Assumes both keys and values are hashable. True in practice.
    """

    def __init__(self, map: Mapping[K, V] | None = None) -> None:
        """
        Make an instance from another dictionary.
        This object will be mutable until it is used as a key.
        """
        super().__init__(map or {})
        self.__hash: int | None = None

    def __repr__(self):
        return f"""{self.__class__.__name__}({
            ', '.join(f'{k!r}={v!r}' for k, v in self.items())
        })"""

    def __hash__(self):
        if self.__hash is None:
            self.__hash = hash(frozenset(self.items()))
        return self.__hash

    def _validate_update(self) -> None:
        if self.__hash is not None:
            raise _LoopIndexError(self)

    def __setitem__(self, key: K, value: V) -> None:
        self._validate_update()
        super().__setitem__(key, value)

    def __delitem__(self, key: K) -> None:
        self._validate_update()
        super().__delitem__(key)

    def clear(self) -> None:
        self._validate_update()
        super().clear()

    def pop(self, *args, **kwargs) -> V:
        self._validate_update()
        return super().pop(*args, **kwargs)

    def popitem(self) -> tuple[K, V]:
        self._validate_update()
        return super().popitem()

    def setdefault(self, key: K, default: V) -> V:
        self._validate_update()
        return super().setdefault(key, default)

    def update(self, *args, **kwargs) -> None:
        self._validate_update()
        super().update(*args, **kwargs)

    def __add__(self, right: Mapping[K, V]) -> Self:
        result = self.__class__(self)
        result.update(right)
        return result


@dataclass
class LoopCache:
    """Class to store a cache for use in :py:meth:`.Workflow.add_empty_loop` and
    :py:meth:`.WorkflowLoop.add_iterations`. Use :py:meth:`build` to get a new instance.

    Parameters
    ----------
    element_dependents:
        Keys are element IDs, values are dicts whose keys are element IDs that depend on
        the key element ID (via `Element.get_dependent_elements_recursively`), and whose
        values are dicts with keys: `group_names`, which is a tuple of the string group
        names associated with the dependent element's element set.
    elements:
        Keys are element IDs, values are dicts with keys: `input_statuses`,
        `input_sources`, and `task_insert_ID`.
    zeroth_iters:
        Keys are element IDs, values are data associated with the zeroth iteration of that
        element, namely a tuple of iteration ID and `ElementIteration.data_idx`.
    data_idx:
        Keys are element IDs, values are data associated with all iterations of that
        element, namely a dict whose keys are the iteration loop index as a tuple, and
        whose values are data indices via `ElementIteration.get_data_idx()`.
    iterations:
        Keys are iteration IDs, values are tuples of element ID and iteration index within
        that element.
    task_iterations:
        Keys are task insert IDs, values are list of all iteration IDs associated with
        that task.
    """

    #: Keys are element IDs, values are dicts whose keys are element IDs that depend on
    #: the key element ID (via `Element.get_dependent_elements_recursively`), and whose
    #: values are dicts with keys: `group_names`, which is a tuple of the string group
    #: names associated with the dependent element's element set.
    element_dependents: dict[int, dict[int, DependentDescriptor]]
    #: Keys are element IDs, values are dicts with keys: `input_statuses`,
    #: `input_sources`, and `task_insert_ID`.
    elements: dict[int, ElementDescriptor]
    #: Keys are element IDs, values are data associated with the zeroth iteration of that
    #: element, namely a tuple of iteration ID and `ElementIteration.data_idx`.
    zeroth_iters: dict[int, tuple[int, DataIndex]]
    #: Keys are element IDs, values are data associated with all iterations of that
    #: element, namely a dict whose keys are the iteration loop index as a tuple, and
    #: whose values are data indices via `ElementIteration.get_data_idx()`.
    data_idx: dict[int, dict[LoopIndex[str, int], DataIndex]]
    #: Keys are iteration IDs, values are tuples of element ID and iteration index within
    #: that element.
    iterations: dict[int, tuple[int, int]]
    #: Keys are task insert IDs, values are list of all iteration IDs associated with
    #: that task.
    task_iterations: dict[int, list[int]]

    @TimeIt.decorator
    def get_iter_IDs(self, loop: Loop) -> list[int]:
        """Retrieve a list of iteration IDs belonging to a given loop."""
        return [
            i_id for t_id in loop.task_insert_IDs for i_id in self.task_iterations[t_id]
        ]

    @TimeIt.decorator
    def get_iter_loop_indices(self, iter_IDs: list[int]) -> Sequence[Mapping[str, int]]:
        """
        Retrieve the mapping from element to loop index for each given iteration.
        """
        iter_loop_idx: list[LoopIndex[str, int]] = []
        for id_ in iter_IDs:
            elem_id, idx = self.iterations[id_]
            iter_loop_idx.append(nth_key(self.data_idx[elem_id], idx))
        return iter_loop_idx

    @TimeIt.decorator
    def update_loop_indices(self, new_loop_name: str, iter_IDs: list[int]) -> None:
        """
        Set the loop indices for a named loop to the given list of iteration IDs.
        """
        elem_ids = {e_ids[0] for k, e_ids in self.iterations.items() if k in iter_IDs}
        new_loop_entry = {new_loop_name: 0}
        for id_ in elem_ids:
            self.data_idx[id_] = {
                k + new_loop_entry: v for k, v in self.data_idx[id_].items()
            }

    @TimeIt.decorator
    def add_iteration(
        self,
        iter_ID: int,
        task_insert_ID: int,
        element_ID: int,
        loop_idx: LoopIndex[str, int],
        data_idx: DataIndex,
    ):
        """Update the cache to include a newly added iteration."""
        self.task_iterations[task_insert_ID].append(iter_ID)
        new_iter_idx = len(self.data_idx[element_ID])
        self.data_idx[element_ID][loop_idx] = data_idx
        self.iterations[iter_ID] = (element_ID, new_iter_idx)

    @classmethod
    @TimeIt.decorator
    def build(cls, workflow: Workflow, loops: list[Loop] | None = None) -> Self:
        """Build a cache of data for use in adding loops and iterations."""

        deps_cache = ObjectCache.build(workflow, dependencies=True, elements=True)

        loops = [*workflow.template.loops, *(loops or ())]
        task_iIDs = {t_id for loop in loops for t_id in loop.task_insert_IDs}
        tasks: list[WorkflowTask] = [
            workflow.tasks.get(insert_ID=t_id) for t_id in sorted(task_iIDs)
        ]
        elem_deps: dict[int, dict[int, DependentDescriptor]] = {}

        # keys: element IDs, values: dict with keys: tuple(loop_idx), values: data index
        data_idx_cache: dict[int, dict[LoopIndex[str, int], DataIndex]] = {}

        # keys: iteration IDs, values: tuple of (element ID, integer index into values
        # dict in `data_idx_cache` [accessed via `.keys()[index]`])
        iters: dict[int, tuple[int, int]] = {}

        # keys: element IDs, values: dict with keys: "input_statues", "input_sources",
        # "task_insert_ID":
        elements: dict[int, ElementDescriptor] = {}

        zeroth_iters: dict[int, tuple[int, DataIndex]] = {}
        task_iterations = defaultdict(list)
        for task in tasks:
            for elem_id in task.element_IDs:
                element = deps_cache.elements[elem_id]
                inp_statuses = task.template.get_input_statuses(element.element_set)
                elements[element.id_] = {
                    "input_statuses": inp_statuses,
                    "input_sources": element.input_sources,
                    "task_insert_ID": task.insert_ID,
                }
                elem_deps[element.id_] = {
                    de_id: {
                        "group_names": tuple(
                            grp.name
                            for grp in deps_cache.elements[de_id].element_set.groups
                        ),
                    }
                    for de_id in deps_cache.elem_elem_dependents_rec[element.id_]
                }
                elem_iters: dict[LoopIndex[str, int], DataIndex] = {}
                for idx, iter_i in enumerate(element.iterations):
                    if idx == 0:
                        zeroth_iters[element.id_] = (iter_i.id_, iter_i.data_idx)
                    elem_iters[iter_i.loop_idx] = iter_i.get_data_idx()
                    task_iterations[task.insert_ID].append(iter_i.id_)
                    iters[iter_i.id_] = (element.id_, idx)
                data_idx_cache[element.id_] = elem_iters

        task_iterations.default_factory = None
        return cls(
            element_dependents=elem_deps,
            elements=elements,
            zeroth_iters=zeroth_iters,
            data_idx=data_idx_cache,
            iterations=iters,
            task_iterations=task_iterations,
        )



"""
Common type aliases.
"""

from __future__ import annotations
from dataclasses import InitVar
from typing import Any, ClassVar, Final, TypeVar, cast, TYPE_CHECKING
from typing_extensions import NotRequired, TypeAlias, TypedDict
from pathlib import Path
import re

if TYPE_CHECKING:
    from collections.abc import Mapping
    from datetime import datetime
    from rich.status import Status
    from .core.object_list import (
        CommandFilesList,
        EnvironmentsList,
        ParametersList,
        TaskSchemasList,
    )
    from .submission.enums import JobscriptElementState
    from .submission.submission import Submission


#: Type of a value that can be treated as a path.
PathLike: TypeAlias = "str | Path | None"


class ParamSource(TypedDict):
    """
    A parameter source descriptor.
    """

    #: Parameter type name.
    type: NotRequired[str]
    #: EAR ID.
    EAR_ID: NotRequired[int]
    #: Task insertion ID.
    task_insert_ID: NotRequired[int]
    #: Action index.
    action_idx: NotRequired[int]
    #: Element index.
    element_idx: NotRequired[int]
    #: Element set index.
    element_set_idx: NotRequired[int]
    #: Element action run index.
    run_idx: NotRequired[int]
    #: Sequence index.
    sequence_idx: NotRequired[int]
    #: Task index.
    task_idx: NotRequired[int]
    #: Name of method used to create the parameter's value(s).
    value_class_method: NotRequired[str]


class KnownSubmission(TypedDict):
    """
    Describes a known submission.
    """

    #: Local ID.
    local_id: int
    #: Workflow global ID.
    workflow_id: str
    #: Whether the submission is active.
    is_active: bool
    #: Submission index.
    sub_idx: int
    #: Submission time.
    submit_time: str
    #: Path to submission.
    path: str
    #: Start time.
    start_time: str
    #: Finish time.
    end_time: str


class KnownSubmissionItem(TypedDict):
    """
    Describes a known submission.
    """

    #: Local ID.
    local_id: int
    #: Workflow global ID.
    workflow_id: str
    #: Path to the workflow.
    workflow_path: str
    #: Time of submission.
    submit_time: str
    #: Parsed time of submission.
    submit_time_obj: NotRequired[datetime | None]
    #: Time of start.
    start_time: str
    #: Parsed time of start.
    start_time_obj: datetime | None
    #: Time of finish.
    end_time: str
    #: Parsed time of finish.
    end_time_obj: datetime | None
    #: Submission index.
    sub_idx: int
    #: Jobscripts in submission.
    jobscripts: list[int]
    #: Active jobscript state.
    active_jobscripts: Mapping[int, Mapping[int, Mapping[int, JobscriptElementState]]]
    #: Whether this is deleted.
    deleted: bool
    #: Whether this is unloadable.
    unloadable: bool
    #: Expanded submission object.
    submission: NotRequired[Submission]


class TemplateComponents(TypedDict):
    """
    Components loaded from templates.
    """

    #: Parameters loaded from templates.
    parameters: NotRequired[ParametersList]
    #: Command files loaded from templates.
    command_files: NotRequired[CommandFilesList]
    #: Execution environments loaded from templates.
    environments: NotRequired[EnvironmentsList]
    #: Task schemas loaded from templates.
    task_schemas: NotRequired[TaskSchemasList]
    #: Scripts discovered by templates.
    scripts: NotRequired[dict[str, Path]]
    #: Programs discovered by templates.
    programs: NotRequired[dict[str, Path]]
    #: Jinja templates discovered by templates.
    jinja_templates: NotRequired[dict[str, Path]]


class MakeWorkflowCommonArgs(TypedDict):
    """
    Common keys used in workflow construction in :py:meth:`BaseApp._make_workflow`.
    """

    path: str | None
    name: str | None
    overwrite: bool
    store: str
    ts_fmt: str | None
    ts_name_fmt: str | None
    store_kwargs: dict[str, Any] | None
    variables: dict[str, Any] | None
    status: Status | None


#: Simplification of :class:`TemplateComponents` to allow some types of
#: internal manipulations.
BasicTemplateComponents: TypeAlias = "dict[str, list[dict]]"


DataIndex: TypeAlias = "dict[str, int | list[int]]"
"""
The type of indices to data. These are *normally* dictionaries of integers,
but can have leaves being lists of integers when dealing with element groups
(i.e., when a downstream element uses outputs from multiple upstream elements,
rather than just a single upstream element).
"""


_T = TypeVar("_T")

_CLASS_VAR_RE: Final = re.compile(r"ClassVar\[(.*)\]")
_INIT_VAR_RE: Final = re.compile(r"InitVar\[(.*)\]")


def hydrate(cls: type[_T]) -> type[_T]:
    """
    Partially hydrates the annotations on fields in a class, so that a @dataclass
    annotation can recognise that ClassVar-annotated fields are class variables.
    """
    anns: dict[str, Any] = {}
    for f, a in cls.__annotations__.items():
        if isinstance(a, str):
            m = _CLASS_VAR_RE.match(a)
            if m:
                anns[f] = cast(Any, ClassVar[m[1]])
                continue
            m = _INIT_VAR_RE.match(a)
            if m:
                anns[f] = cast(Any, InitVar(cast(type, m[1])))
                continue
        anns[f] = a
    cls.__annotations__ = anns
    return cls



import os
from textwrap import dedent
import hpcflow.app as hf

import pytest


@pytest.mark.integration
def test_basic_jinja_template(null_config, tmp_path):
    jinja_template_name = "test_template.txt"
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[
            hf.SchemaInput(parameter=hf.Parameter("name")),
            hf.SchemaInput(parameter=hf.Parameter("fruits")),
        ],
        actions=[hf.Action(jinja_template=f"test/{jinja_template_name}")],
    )
    t1 = hf.Task(schema=s1, inputs={"name": "George", "fruits": ["apple", "orange"]})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="jinja_template_test",
        path=tmp_path,
    )
    wk.submit(wait=True, add_to_known=False, status=False)
    run_dir = wk.get_all_EARs()[0].get_directory()
    rendered = run_dir.joinpath(jinja_template_name)

    expected = dedent(
        """\
        Hola, George!

        This is a template, with a loop. Here are your specified fruits:
        - apple
        - orange
        """
    )

    assert rendered.is_file()
    assert rendered.read_text() == expected


@pytest.mark.integration
def test_jinja_template_path_with_resource_var(null_config, tmp_path):
    jinja_template_name = "test_template.txt"
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[
            hf.SchemaInput(parameter=hf.Parameter("name")),
            hf.SchemaInput(parameter=hf.Parameter("fruits")),
        ],
        actions=[
            hf.Action(jinja_template=f"<<resource:resources_id>>/{jinja_template_name}")
        ],
    )
    t1 = hf.Task(schema=s1, inputs={"name": "George", "fruits": ["apple", "orange"]})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="jinja_template_test",
        path=tmp_path,
        resources={"any": {"resources_id": "test"}},
    )
    wk.submit(wait=True, add_to_known=False, status=False)
    run_dir = wk.get_all_EARs()[0].get_directory()
    rendered = run_dir.joinpath(jinja_template_name)

    expected = dedent(
        """\
        Hola, George!

        This is a template, with a loop. Here are your specified fruits:
        - apple
        - orange
        """
    )

    assert rendered.is_file()
    assert rendered.read_text() == expected


@pytest.mark.integration
def test_builtin_jinja_template_path_with_env_var(new_null_config, tmp_path):
    env = hf.Environment(
        name="null_env",
        specifiers={"key": "test"},  # reference this specifier in the jinja template path
        executables=[],
    )
    hf.envs.add_object(env, skip_duplicates=True)

    jinja_template_name = "test_template.txt"
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[
            hf.SchemaInput(parameter=hf.Parameter("name")),
            hf.SchemaInput(parameter=hf.Parameter("fruits")),
        ],
        actions=[hf.Action(jinja_template=f"<<env:key>>/{jinja_template_name}")],
    )
    t1 = hf.Task(schema=s1, inputs={"name": "George", "fruits": ["apple", "orange"]})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="jinja_template_test",
        path=tmp_path,
        environments={"null_env": {"key": "test"}},
    )
    wk.submit(wait=True, add_to_known=False, status=False)
    run_dir = wk.get_all_EARs()[0].get_directory()
    rendered = run_dir.joinpath(jinja_template_name)

    expected = dedent(
        """\
        Hola, George!

        This is a template, with a loop. Here are your specified fruits:
        - apple
        - orange
        """
    )

    assert rendered.is_file()
    assert rendered.read_text() == expected

    hf.reload_template_components()  # remove extra env


@pytest.mark.integration
def test_builtin_jinja_template_path_with_param_var(new_null_config, tmp_path):
    jinja_template_name = "test_template.txt"
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[
            hf.SchemaInput(parameter=hf.Parameter("name")),
            hf.SchemaInput(parameter=hf.Parameter("fruits")),
            hf.SchemaInput(parameter=hf.Parameter("key")),
        ],
        actions=[hf.Action(jinja_template=f"<<parameter:key>>/{jinja_template_name}")],
    )
    t1 = hf.Task(
        schema=s1, inputs={"name": "George", "fruits": ["apple", "orange"], "key": "test"}
    )
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="jinja_template_test",
        path=tmp_path,
    )
    wk.submit(wait=True, add_to_known=False, status=False)
    run_dir = wk.get_all_EARs()[0].get_directory()
    rendered = run_dir.joinpath(jinja_template_name)

    expected = dedent(
        """\
        Hola, George!

        This is a template, with a loop. Here are your specified fruits:
        - apple
        - orange
        """
    )

    assert rendered.is_file()
    assert rendered.read_text() == expected



from pathlib import Path
from textwrap import dedent
import pytest
import zarr  # type: ignore
import numpy as np
from numpy.typing import NDArray
from hpcflow.sdk.core.errors import (
    InvalidIdentifier,
    MissingVariableSubstitutionError,
    YAMLError,
)

from hpcflow.sdk.core.utils import (
    JSONLikeDirSnapShot,
    bisect_slice,
    dict_values_process_flat,
    flatten,
    get_nested_indices,
    is_fsspec_url,
    linspace_rect,
    nth_key,
    nth_value,
    process_string_nodes,
    read_YAML_str,
    replace_items,
    check_valid_py_identifier,
    reshape,
    split_param_label,
    substitute_string_vars,
    swap_nested_dict_keys,
)


def test_get_nested_indices_expected_values_size_2_nest_levels_2():
    size, nest_levels = (2, 2)
    assert [
        get_nested_indices(i, size=size, nest_levels=nest_levels)
        for i in range(size**nest_levels)
    ] == [
        [0, 0],
        [0, 1],
        [1, 0],
        [1, 1],
    ]


def test_get_nested_indices_expected_values_size_2_nest_levels_4():
    size, nest_levels = (2, 4)
    assert [
        get_nested_indices(i, size=size, nest_levels=nest_levels)
        for i in range(size**nest_levels)
    ] == [
        [0, 0, 0, 0],
        [0, 0, 0, 1],
        [0, 0, 1, 0],
        [0, 0, 1, 1],
        [0, 1, 0, 0],
        [0, 1, 0, 1],
        [0, 1, 1, 0],
        [0, 1, 1, 1],
        [1, 0, 0, 0],
        [1, 0, 0, 1],
        [1, 0, 1, 0],
        [1, 0, 1, 1],
        [1, 1, 0, 0],
        [1, 1, 0, 1],
        [1, 1, 1, 0],
        [1, 1, 1, 1],
    ]


def test_get_nested_indices_expected_values_size_4_nest_levels_2():
    size, nest_levels = (4, 2)
    assert [
        get_nested_indices(i, size=size, nest_levels=nest_levels)
        for i in range(size**nest_levels)
    ] == [
        [0, 0],
        [0, 1],
        [0, 2],
        [0, 3],
        [1, 0],
        [1, 1],
        [1, 2],
        [1, 3],
        [2, 0],
        [2, 1],
        [2, 2],
        [2, 3],
        [3, 0],
        [3, 1],
        [3, 2],
        [3, 3],
    ]


def test_get_nested_indices_expected_values_size_4_nest_levels_3():
    size, nest_levels = (4, 3)
    assert [
        get_nested_indices(i, size=size, nest_levels=nest_levels)
        for i in range(size**nest_levels)
    ] == [
        [0, 0, 0],
        [0, 0, 1],
        [0, 0, 2],
        [0, 0, 3],
        [0, 1, 0],
        [0, 1, 1],
        [0, 1, 2],
        [0, 1, 3],
        [0, 2, 0],
        [0, 2, 1],
        [0, 2, 2],
        [0, 2, 3],
        [0, 3, 0],
        [0, 3, 1],
        [0, 3, 2],
        [0, 3, 3],
        [1, 0, 0],
        [1, 0, 1],
        [1, 0, 2],
        [1, 0, 3],
        [1, 1, 0],
        [1, 1, 1],
        [1, 1, 2],
        [1, 1, 3],
        [1, 2, 0],
        [1, 2, 1],
        [1, 2, 2],
        [1, 2, 3],
        [1, 3, 0],
        [1, 3, 1],
        [1, 3, 2],
        [1, 3, 3],
        [2, 0, 0],
        [2, 0, 1],
        [2, 0, 2],
        [2, 0, 3],
        [2, 1, 0],
        [2, 1, 1],
        [2, 1, 2],
        [2, 1, 3],
        [2, 2, 0],
        [2, 2, 1],
        [2, 2, 2],
        [2, 2, 3],
        [2, 3, 0],
        [2, 3, 1],
        [2, 3, 2],
        [2, 3, 3],
        [3, 0, 0],
        [3, 0, 1],
        [3, 0, 2],
        [3, 0, 3],
        [3, 1, 0],
        [3, 1, 1],
        [3, 1, 2],
        [3, 1, 3],
        [3, 2, 0],
        [3, 2, 1],
        [3, 2, 2],
        [3, 2, 3],
        [3, 3, 0],
        [3, 3, 1],
        [3, 3, 2],
        [3, 3, 3],
    ]


def test_get_nest_index_raise_on_rollover():
    size = 4
    nest_levels = 3
    with pytest.raises(ValueError):
        get_nested_indices(
            idx=size**nest_levels,
            size=size,
            nest_levels=nest_levels,
            raise_on_rollover=True,
        )


@pytest.fixture
def zarr_column_array(tmp_path: Path):
    headers = ["a", "b", "c"]
    num_rows = 2
    fill_value = -1
    arr = zarr.open_array(
        store=f"{tmp_path}/zarr_column_array_test.zarr",
        mode="w",
        shape=(num_rows, len(headers)),
        dtype=int,
        fill_value=fill_value,
    )
    arr[:] = np.arange(np.prod(arr.shape)).reshape(arr.shape)
    return arr, headers, fill_value


def test_bisect_slice():
    tot_len = 8
    tot_lst = list(range(tot_len))
    for sel_start in range(tot_len + 1):
        for sel_step in range(1, tot_len):
            for sel_stop in range(sel_start, tot_len + 1):
                for len_A in range(tot_len):
                    lst_A = tot_lst[:len_A]
                    lst_B = tot_lst[len_A:]
                    selection = slice(sel_start, sel_stop, sel_step)
                    slice_A, slice_B = bisect_slice(selection, len(lst_A))
                    sub_A = lst_A[slice_A]
                    sub_B = lst_B[slice_B]
                    assert sub_A + sub_B == tot_lst[selection]


def test_replace_items():
    lst = [0, 1, 2, 3]
    ins = [10, 11]
    assert replace_items(lst, start=1, end=3, repl=ins) == [0, 10, 11, 3]


def test_replace_items_single_item():
    lst = [0]
    ins = [10, 11]
    assert replace_items(lst, start=0, end=1, repl=ins) == [10, 11]


def test_raise_check_valid_py_identifier_empty_str():
    with pytest.raises(InvalidIdentifier):
        check_valid_py_identifier("")


def test_raise_check_valid_py_identifier_start_digit():
    with pytest.raises(InvalidIdentifier):
        check_valid_py_identifier("9sdj")


def test_raise_check_valid_py_identifier_single_digit():
    with pytest.raises(InvalidIdentifier):
        check_valid_py_identifier("9")


def test_raise_check_valid_py_identifier_py_keyword():
    with pytest.raises(InvalidIdentifier):
        check_valid_py_identifier("if")


def test_raise_check_valid_py_identifier_non_str():
    with pytest.raises(InvalidIdentifier):
        check_valid_py_identifier(0.123)


def test_raise_check_valid_py_identifier_starts_underscore():
    with pytest.raises(InvalidIdentifier):
        check_valid_py_identifier("_test")


def test_expected_return_check_valid_py_identifier_internal_underscore():
    assert check_valid_py_identifier("test_ok") == "test_ok"


def test_expected_return_check_valid_py_identifier_end_underscore():
    assert check_valid_py_identifier("test_ok_") == "test_ok_"


def test_expected_return_check_valid_py_identifier_all_latin_alpha():
    assert check_valid_py_identifier("abc") == "abc"


def test_expected_return_check_valid_py_identifier_all_latin_alphanumeric():
    assert check_valid_py_identifier("abc123") == "abc123"


def test_expected_return_check_valid_py_identifier_all_greek_alpha():
    assert check_valid_py_identifier("αβγ") == "αβγ"


def test_flatten_reshape_round_trip_depth_2():
    lst = [[[1, 2], [3]], [[4, 5, 6], [7, 8], [9, 10]]]
    lst_flat, lens = flatten(lst)
    assert lst_flat == list(range(1, 11)) and reshape(lst_flat, lens) == lst


def test_flatten_reshape_round_trip_depth_0():
    lst = [1, 2, 3]
    lst_flat, lens = flatten(lst)
    assert lst_flat == list(range(1, 4)) and reshape(lst_flat, lens) == lst


def test_flatten_reshape_round_trip_depth_1():
    lst = [[4, 5, 6], [7, 8], [9, 10]]
    lst_flat, lens = flatten(lst)
    assert lst_flat == list(range(4, 11)) and reshape(lst_flat, lens) == lst


def test_flatten_expected_return_first_item_empty_list():
    lst = [[], [1]]
    flt = flatten(lst)
    assert flt == ([1], ([0, 1],))


def test_flatten_expected_return_second_item_empty_list():
    lst = [[1], []]
    flt = flatten(lst)
    assert flt == ([1], ([1, 0],))


def test_is_fsspec_url_simple():
    assert is_fsspec_url("github://dask:fastparquet@main/test-data/nation.csv")


def test_is_fsspec_url_compound():
    assert is_fsspec_url("dask::s3://bucket/key")


def test_is_fsspec_url_compound_complex():
    assert is_fsspec_url("simplecache::zip://*.csv::gcs://bucket/afile.zip")


def test_is_fsspec_url_false_cwd():
    assert not is_fsspec_url(".")


def test_is_fsspec_url_false_local_win_abs_path():
    assert not is_fsspec_url("C:\\my_files")


def test_is_fsspec_url_false_local_win_rel_path():
    assert not is_fsspec_url(".\\a\\b\\c")


def test_is_fsspec_url_false_local_win_rel_up_path():
    assert not is_fsspec_url("..\\a\\b\\c")


def test_is_fsspec_url_false_local_nix_abs_path():
    assert not is_fsspec_url("/mnt/c/my_files")


def test_is_fsspec_url_false_local_nix_rel_path():
    assert not is_fsspec_url("./a/b/c")


def test_is_fsspec_url_false_local_nix_rel_up_path():
    assert not is_fsspec_url("../a/b/c")


def test_JSONLikeDirSnapShot_round_trip(tmpdir):
    snap_0 = JSONLikeDirSnapShot()
    snap_0.take(str(tmpdir))
    js_0 = snap_0.to_json_like()
    snap_0_rl = JSONLikeDirSnapShot(**js_0)
    assert snap_0._stat_info == snap_0_rl._stat_info


def test_split_param_label(null_config):
    assert split_param_label("inputs.p1") == ("inputs.p1", None)
    assert split_param_label("inputs.p1[one]") == ("inputs.p1", "one")
    assert split_param_label("p1") == ("p1", None)
    assert split_param_label("p1[one]") == ("p1", "one")
    assert split_param_label("p1.sub.data") == ("p1.sub.data", None)
    assert split_param_label("p1.sub.data[one]") == ("p1.sub.data", "one")


def test_process_string_nodes(null_config):
    str_processor = str.upper
    data = {
        "a": [1, 2, 3],
        "b": "hello",
        "c": {
            "d": "hi",
            "e": {"s1", "S2", "3s"},
            "f": ("abc", "def"),
        },
    }
    assert process_string_nodes(data, str_processor) == {
        "a": [1, 2, 3],
        "b": "HELLO",
        "c": {
            "d": "HI",
            "e": {"S1", "S2", "3S"},
            "f": ("ABC", "DEF"),
        },
    }


def test_linspace_rect_endpoint_true():
    rect = linspace_rect(start=[0, 0], stop=[1, 10], num=[3, 3], endpoint=True)
    expected = np.array(
        [
            [0.0, 10.0],
            [0.5, 10.0],
            [1.0, 10.0],
            [1.0, 5.0],
            [0.0, 0.0],
            [0.5, 0.0],
            [1.0, 0.0],
            [0.0, 5.0],
        ]
    ).T
    assert np.allclose(rect, expected)


def test_linspace_rect_endpoint_false():
    rect = linspace_rect(start=[0, 0], stop=[1, 10], num=[3, 3], endpoint=False)
    expected = np.array(
        [
            [0.0, 20 / 3],
            [1 / 3, 20 / 3],
            [2 / 3, 20 / 3],
            [2 / 3, 10 / 3],
            [0.0, 0.0],
            [1 / 3, 0.0],
            [2 / 3, 0.0],
            [0.0, 10 / 3],
        ]
    ).T
    assert np.allclose(rect, expected)


def test_linspace_rect_endpoint_true_include_left_right():
    rect = linspace_rect(
        start=[0, 0],
        stop=[1, 5],
        num=[4, 3],
        include=["left", "right"],
    )
    expected = np.array(
        [[1.0, 0.0], [1.0, 2.5], [1.0, 5.0], [0.0, 0.0], [0.0, 2.5], [0.0, 5.0]]
    ).T
    assert np.allclose(rect, expected)


def test_linspace_rect_endpoint_true_include_top_bottom():
    rect = linspace_rect(
        start=[0, 0],
        stop=[1, 5],
        num=[4, 3],
        include=["top", "bottom"],
    )
    expected = np.array(
        [
            [0.0, 5.0],
            [0.33333333, 5.0],
            [0.66666667, 5.0],
            [1.0, 5.0],
            [0.0, 0.0],
            [0.33333333, 0.0],
            [0.66666667, 0.0],
            [1.0, 0.0],
        ]
    ).T
    assert np.allclose(rect, expected)


def test_linspace_rect_endpoint_true_include_top_right():
    rect = linspace_rect(
        start=[0, 0],
        stop=[1, 5],
        num=[4, 3],
        include=["top", "right"],
    )
    expected = np.array(
        [
            [0.0, 5.0],
            [0.33333333, 5.0],
            [0.66666667, 5.0],
            [1.0, 5.0],
            [1.0, 0.0],
            [1.0, 2.5],
        ]
    ).T
    assert np.allclose(rect, expected)


def test_swap_nested_dict_keys():
    dct = {
        "p1": {"format": "direct", "all_iterations": True},
        "p2": {"format": "json"},
        "p3": {"format": "direct"},
    }
    assert swap_nested_dict_keys(dct, inner_key="format") == {
        "direct": {"p1": {"all_iterations": True}, "p3": {}},
        "json": {"p2": {}},
    }


def test_substitute_string_vars():
    assert (
        substitute_string_vars(
            "hello <<var:my_name>>!",
            variables={"my_name": "bob"},
        )
        == "hello bob!"
    )


def test_substitute_string_vars_repeated_var():
    assert (
        substitute_string_vars(
            "hello <<var:my_name>>; how are you <<var:my_name>>!",
            variables={"my_name": "bob"},
        )
        == "hello bob; how are you bob!"
    )


def test_substitute_string_vars_empty_vars():
    assert substitute_string_vars("hello bob!", variables={}) == "hello bob!"


def test_substitute_string_vars_raise_no_vars():
    with pytest.raises(MissingVariableSubstitutionError):
        substitute_string_vars("hello <<var:my_name>>", variables={})


def test_substitute_string_vars_raise_missing():
    with pytest.raises(MissingVariableSubstitutionError):
        substitute_string_vars("hello <<var:my_name>>", variables={"a": "b"})


def test_substitute_string_vars_non_str():
    assert substitute_string_vars("<<var:a>>", variables={"a": 2}) == "2"


def test_substitute_string_vars_default_value():
    assert (
        substitute_string_vars("hello <<var:my_name[default=bill]>>!", variables={})
        == "hello bill!"
    )


def test_substitute_string_vars_default_value_with_specified():
    assert (
        substitute_string_vars(
            "hello <<var:my_name[default=bill]>>!", variables={"my_name": "bob"}
        )
        == "hello bob!"
    )


def test_dict_values_process_flat():
    d = {"a": 0, "b": [1, 2], "c": 5}
    assert dict_values_process_flat(d, callable=lambda x: [i + 3 for i in x]) == {
        "a": 3,
        "b": [4, 5],
        "c": 8,
    }


def test_dict_values_process_flat_no_lists():
    d = {"a": 0, "b": 1, "c": 2}
    assert dict_values_process_flat(d, callable=lambda x: [i + 3 for i in x]) == {
        "a": 3,
        "b": 4,
        "c": 5,
    }


def test_dict_values_process_flat_single_item_lists():
    d = {"a": [0], "b": [1], "c": [2]}
    assert dict_values_process_flat(d, callable=lambda x: [i + 3 for i in x]) == {
        "a": [3],
        "b": [4],
        "c": [5],
    }


def test_nth_key():
    dct = {"a": 1, "b": 2}
    assert [nth_key(dct, i) for i in range(len(dct))] == ["a", "b"]


def test_nth_value():
    dct = {"a": 1, "b": 2}
    assert [nth_value(dct, i) for i in range(len(dct))] == [1, 2]


def test_nth_key_raises():
    dct = {"a": 1, "b": 2}
    with pytest.raises(Exception):
        nth_key(dct, 2)

    with pytest.raises(Exception):
        nth_key(dct, -1)


def test_read_YAML_str():
    good_yaml = dedent(
        """\
        a: 1
        b: 2
        """
    )
    assert read_YAML_str(good_yaml) == {"a": 1, "b": 2}


def test_read_YAML_str_raise_on_bad_indent():
    bad_yaml = dedent(
        """\
        a: 1
          b: 2
        """
    )
    with pytest.raises(YAMLError):
        read_YAML_str(bad_yaml)


def test_read_YAML_str_raise_on_mixed_tabs_spaces():
    bad_yaml = dedent(
        """\
        a:
          a1: 2 # this has a space indent
        	a2: 3 # this has a tab indent
        """
    )
    with pytest.raises(YAMLError):
        read_YAML_str(bad_yaml)



"""
Types for the submission subsystem.
"""

from __future__ import annotations
from typing import Any, TYPE_CHECKING
from typing_extensions import NotRequired, TypeAlias, TypedDict

if TYPE_CHECKING:
    from collections.abc import Sequence
    from datetime import datetime
    from numpy.typing import NDArray
    from ..core.element import ElementResources


class JobScriptDescriptor(TypedDict):
    """
    Descriptor for a jobscript.
    """

    #: Resources required by the jobscript.
    resources: Any
    #: Elements handled by the jobscript.
    elements: dict[int, list[int]]
    #: Dependencies of the jobscript.
    dependencies: NotRequired[dict[int, ResolvedJobscriptBlockDependencies]]
    #: Hash of resources.
    resource_hash: NotRequired[str]


class ResolvedJobscriptBlockDependencies(TypedDict):
    """
    The resolution of a jobscript block dependency. This represents the dependency of one
    jobscript block on another.
    """

    #: Mapping of jobscript elements.
    js_element_mapping: dict[int, list[int]]
    #: Whether this is an array mapping.
    is_array: NotRequired[bool]


class JobScriptCreationArguments(TypedDict):
    """
    Arguments to pass to create a :class:`Jobscript`.
    """

    # TODO: this currently represents a mix of arguments for both jobscripts and jobscript
    # blocks; need to separate

    #: The task insertion IDs.
    task_insert_IDs: list[int]
    #: The actions of the tasks.
    task_actions: list[tuple[int, int, int]]
    #: The elements of the tasks.
    task_elements: dict[int, list[int]]
    #: Element action run information.
    EAR_ID: NDArray
    #: Resources to use.
    resources: NotRequired[ElementResources]
    #: Description of what loops are in play.
    task_loop_idx: list[dict[str, int]]
    #: Description of dependencies.
    dependencies: dict[int | tuple[int, int], ResolvedJobscriptBlockDependencies]
    #: Whether this is an array jobscript.
    is_array: NotRequired[bool]
    #: When the jobscript was submitted, if known.
    submit_time: NotRequired[datetime]
    #: Where the jobscript was submitted, if known.
    submit_hostname: NotRequired[str]
    #: Description of what the jobscript was submitted to, if known.
    submit_machine: NotRequired[str]
    #: The command line used to do the commit, if known.
    submit_cmdline: NotRequired[list[str]]
    #: The job ID from the scheduler, if known.
    scheduler_job_ID: NotRequired[str]
    #: The process ID of the subprocess, if known.
    process_ID: NotRequired[int]
    #: Version info about the target system.
    version_info: NotRequired[dict[str, str | list[str]]]
    #: The name of the OS.
    os_name: NotRequired[str]
    #: The name of the shell.
    shell_name: NotRequired[str]
    #: The scheduler used.
    scheduler_name: NotRequired[str]
    #: Whether the jobscript is currently running.
    running: NotRequired[bool]
    #: Do not supply!
    resource_hash: NotRequired[str]
    #: Do not supply!
    elements: NotRequired[dict[int, list[int]]]


class SchedulerRef(TypedDict):
    """
    Scheduler reference descriptor.
    """

    #: Jobscript references.
    js_refs: list  # Internal type is horrible and variable
    #: Number of jobscript elements.
    num_js_elements: int


class SubmissionPart(TypedDict):
    """
    A part of a submission.
    """

    #: Timestamp for when this part was submitted.
    submit_time: datetime
    #: The jobscripts involved in this submission.
    jobscripts: list[int]


# This needs PEP 728 for a better type, alas
#: Version data.
VersionInfo: TypeAlias = "dict[str, str | list[str]]"


# TODO: This really doesn't belong here?!
class JobscriptHeaderArgs(TypedDict):
    """
    Keyword arguments to use when creating a job script from a
    :class:`Jobscript`.
    """

    #: Application invocation. (Arguments, etc.)
    app_invoc: str | Sequence[str]
    #: Workflow application alias.
    workflow_app_alias: NotRequired[str]
    #: Environment setup.
    env_setup: NotRequired[str]
    #: Application name in CAPS
    app_caps: NotRequired[str]
    #: Configuration directory.
    config_dir: NotRequired[str]
    #: Configuration key.
    config_invoc_key: NotRequired[Any]



import sys
import pytest

from hpcflow.app import app as hf
from hpcflow.sdk.core.test_utils import make_test_data_YAML_workflow


@pytest.mark.integration
@pytest.mark.skipif(
    sys.version_info < (3, 9), reason="Python 3.8 support is being removed anyway."
)
def test_workflow_1_zip(tmp_path, new_null_config):
    wk = make_test_data_YAML_workflow("workflow_1.yaml", path=tmp_path)
    wk.submit(wait=True, add_to_known=False, status=False)

    zip_path = wk.zip(path=tmp_path)
    wkz = hf.Workflow(zip_path)
    assert wkz.tasks[0].elements[0].outputs.p2.value == "201"



import os
import time

import pytest
from hpcflow.app import app as hf


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_output_file_parser_parses_file(null_config, tmp_path):
    out_file_name = "my_output_file.txt"
    out_file = hf.FileSpec(label="my_output_file", name=out_file_name)

    if os.name == "nt":
        cmd = f"Set-Content -Path {out_file_name} -Value (<<parameter:p1>> + 100)"
    else:
        cmd = f"echo $(( <<parameter:p1>> + 100 )) > {out_file_name}"

    act = hf.Action(
        commands=[hf.Command(cmd)],
        output_file_parsers=[
            hf.OutputFileParser(
                output_files=[out_file],
                output=hf.Parameter("p2"),
                script="<<script:output_file_parser_basic.py>>",
            ),
        ],
        environments=[hf.ActionEnvironment(environment="python_env")],
    )
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaInput(parameter=hf.Parameter("p2"))],
        actions=[act],
    )

    p1_val = 101
    p2_val_expected = p1_val + 100
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="output_file_parser_test",
        path=tmp_path,
    )

    wk.submit(wait=True, add_to_known=False)

    # check the command successfully generated the output file:
    run_0 = wk.get_all_EARs()[0]
    exec_path = run_0.get_directory()
    out_file_path = exec_path.joinpath(out_file.name.name)
    out_file_contents = out_file_path.read_text()
    assert out_file_contents.strip() == str(p2_val_expected)

    # check the output is parsed correctly:
    assert wk.tasks[0].elements[0].outputs.p2.value == p2_val_expected


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_OFP_std_stream_redirect_on_exception(new_null_config, tmp_path):
    """Test exceptions raised by the app during execution of an OFP script are printed to the
    std-stream redirect file (and not the jobscript's standard error file)."""

    # define a custom python environment which redefines the `WK_PATH` shell variable to
    # a nonsense value so the app cannot load the workflow and thus raises an exception

    app_caps = hf.package_name.upper()
    if os.name == "nt":
        env_cmd = f'$env:{app_caps}_WK_PATH = "nonsense_path"'
    else:
        env_cmd = f'export {app_caps}_WK_PATH="nonsense_path"'

    env_cmd += "; python <<script_path>> <<args>>"
    bad_env = hf.Environment(
        name="bad_python_env",
        executables=[
            hf.Executable(
                label="python_script",
                instances=[
                    hf.ExecutableInstance(
                        command=env_cmd,
                        num_cores=1,
                        parallel_mode=None,
                    )
                ],
            )
        ],
    )
    hf.envs.add_object(bad_env, skip_duplicates=True)

    out_file_name = "my_output_file.txt"
    out_file = hf.FileSpec(label="my_output_file", name=out_file_name)

    if os.name == "nt":
        cmd = f"Set-Content -Path {out_file_name} -Value (<<parameter:p1>> + 100)"
    else:
        cmd = f"echo $(( <<parameter:p1>> + 100 )) > {out_file_name}"

    act = hf.Action(
        commands=[hf.Command(cmd)],
        output_file_parsers=[
            hf.OutputFileParser(
                output_files=[out_file],
                output=hf.Parameter("p2"),
                script="<<script:output_file_parser_basic.py>>",
            ),
        ],
        environments=[hf.ActionEnvironment(environment="bad_python_env")],
    )

    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaInput(parameter=hf.Parameter("p2"))],
        actions=[act],
    )

    p1_val = 101
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="output_file_parser_test",
        path=tmp_path,
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    # jobscript stderr should be empty
    assert not wk.submissions[0].jobscripts[0].direct_stderr_path.read_text()

    # std stream file has workflow not found traceback
    run = wk.get_all_EARs()[1]
    std_stream_path = run.get_app_std_path()
    assert std_stream_path.is_file()
    assert "WorkflowNotFoundError" in std_stream_path.read_text()

    hf.reload_template_components()  # remove extra envs


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_OFP_std_out_std_err_not_redirected(null_config, tmp_path):
    """Test that standard error and output streams from an OFP script are written to the jobscript
    standard error and output files."""
    out_file_name = "my_output_file.txt"
    out_file = hf.FileSpec(label="my_output_file", name=out_file_name)

    if os.name == "nt":
        cmd = f"Set-Content -Path {out_file_name} -Value (<<parameter:p1>> + 100)"
    else:
        cmd = f"echo $(( <<parameter:p1>> + 100 )) > {out_file_name}"

    act = hf.Action(
        commands=[hf.Command(cmd)],
        output_file_parsers=[
            hf.OutputFileParser(
                output_files=[out_file],
                output=hf.Parameter("p2"),
                inputs=["p1"],
                script="<<script:output_file_parser_test_stdout_stderr.py>>",
            ),
        ],
        environments=[hf.ActionEnvironment(environment="python_env")],
    )

    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaInput(parameter=hf.Parameter("p2"))],
        actions=[act],
    )
    p1_val = 101
    stdout_msg = str(p1_val)
    stderr_msg = str(p1_val)
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="ouput_file_parser_test",
        path=tmp_path,
    )
    wk.submit(wait=True, add_to_known=False)

    if wk.submissions[0].jobscripts[0].resources.combine_jobscript_std:
        std_out_err = wk.submissions[0].jobscripts[0].direct_std_out_err_path.read_text()
        assert std_out_err.strip() == f"{stdout_msg}\n{stderr_msg}"
    else:
        std_out = wk.submissions[0].jobscripts[0].direct_stdout_path.read_text()
        std_err = wk.submissions[0].jobscripts[0].direct_stderr_path.read_text()
        assert std_out.strip() == stdout_msg
        assert std_err.strip() == stderr_msg


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_output_file_parser_pass_env_spec(null_config, tmp_path):
    out_file_name = "my_output_file.txt"
    out_file = hf.FileSpec(label="my_output_file", name=out_file_name)

    if os.name == "nt":
        cmd = f"Set-Content -Path {out_file_name} -Value (<<parameter:p1>> + 100)"
    else:
        cmd = f"echo $(( <<parameter:p1>> + 100 )) > {out_file_name}"

    act = hf.Action(
        commands=[hf.Command(cmd)],
        output_file_parsers=[
            hf.OutputFileParser(
                output_files=[out_file],
                output=hf.Parameter("p2"),
                script="<<script:env_specifier_test/output_file_parser_pass_env_spec.py>>",
                script_pass_env_spec=True,
            ),
        ],
        environments=[hf.ActionEnvironment(environment="python_env")],
    )
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaInput(parameter=hf.Parameter("p2"))],
        actions=[act],
    )

    t1 = hf.Task(schema=s1, inputs={"p1": 101})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="output_file_parser_pass_env_spec",
        path=tmp_path,
    )

    wk.submit(wait=True, add_to_known=False, status=False)

    std_out = wk.submissions[0].jobscripts[0].direct_stdout_path.read_text().strip()
    assert std_out == "{'name': 'python_env'}"


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_env_specifier_in_output_file_parser_script_path(new_null_config, tmp_path):

    py_env = hf.Environment(
        name="python_env",
        specifiers={"version": "v1"},
        executables=[
            hf.Executable(
                label="python_script",
                instances=[
                    hf.ExecutableInstance(
                        command="python <<script_path>> <<args>>",
                        num_cores=1,
                        parallel_mode=None,
                    )
                ],
            )
        ],
    )
    hf.envs.add_object(py_env, skip_duplicates=True)

    out_file_name = "my_output_file.txt"
    out_file = hf.FileSpec(label="my_output_file", name=out_file_name)

    if os.name == "nt":
        cmd = f"Set-Content -Path {out_file_name} -Value (<<parameter:p1>> + 100)"
    else:
        cmd = f"echo $(( <<parameter:p1>> + 100 )) > {out_file_name}"

    act = hf.Action(
        commands=[hf.Command(cmd)],
        output_file_parsers=[
            hf.OutputFileParser(
                output_files=[out_file],
                output=hf.Parameter("p2"),
                script="<<script:env_specifier_test/<<env:version>>/output_file_parser_basic.py>>",
            ),
        ],
        environments=[hf.ActionEnvironment(environment="python_env")],
    )
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaInput(parameter=hf.Parameter("p2"))],
        actions=[act],
    )

    p1_val = 101
    p2_val_expected = p1_val + 100
    t1 = hf.Task(
        schema=s1,
        inputs={"p1": p1_val},
        environments={"python_env": {"version": "v1"}},
    )
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="output_file_parser_test_env_specifier",
        path=tmp_path,
    )

    wk.submit(wait=True, add_to_known=False, status=False)

    # check the command successfully generated the output file:
    run_0 = wk.get_all_EARs()[0]
    exec_path = run_0.get_directory()
    out_file_path = exec_path.joinpath(out_file.name.name)
    out_file_contents = out_file_path.read_text()
    assert out_file_contents.strip() == str(p2_val_expected)

    # check the output is parsed correctly:
    assert wk.tasks[0].elements[0].outputs.p2.value == p2_val_expected

    hf.reload_template_components()  # remove extra envs


@pytest.mark.integration
def test_no_script_no_output_saves_files(null_config, tmp_path):
    """Check we can use an output file parser with no script or output to save files."""
    out_file_name = "my_output_file.txt"
    out_file = hf.FileSpec(label="my_output_file", name=out_file_name)

    if os.name == "nt":
        cmd = f"Set-Content -Path {out_file_name} -Value (<<parameter:p1>> + 100)"
    else:
        cmd = f"echo $(( <<parameter:p1>> + 100 )) > {out_file_name}"

    act = hf.Action(
        commands=[hf.Command(cmd)],
        output_file_parsers=[hf.OutputFileParser(output_files=[out_file])],
        environments=[hf.ActionEnvironment(environment="python_env")],
    )
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        actions=[act],
    )

    p1_val = 101
    p2_val_expected = p1_val + 100
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="output_file_parser_test_no_output_no_script",
        path=tmp_path,
    )

    wk.submit(wait=True, add_to_known=False, status=False)

    # check the output file is saved to artifacts:
    run_0 = wk.get_all_EARs()[0]
    exec_path = run_0.get_directory()
    out_file_path = exec_path.joinpath(out_file.name.name)
    out_file_contents = out_file_path.read_text()
    assert out_file_contents.strip() == str(p2_val_expected)

    # check no scripts generated
    assert not any(wk.submissions[0].scripts_path.iterdir())



import h5py  # type: ignore[import-untyped]


def main_script_test_hdf5_in_obj_group(_input_files):
    # read inputs
    with h5py.File(_input_files["hdf5"], mode="r") as fh:
        all_a = [p1c_dat.attrs["a"].item() for p1c_dat in fh["p1c"].values()]

    # process
    p2 = sum(all_a) + 100

    return {"p2": p2}



"""
Model of files in the run directory.
"""

from __future__ import annotations
from typing import Any, TYPE_CHECKING
from hpcflow.sdk.core.app_aware import AppAware
from hpcflow.sdk.core.utils import JSONLikeDirSnapShot

if TYPE_CHECKING:
    from ..submission.shells.base import Shell
    from .types import BlockActionKey


class RunDirAppFiles(AppAware):
    """A class to encapsulate the naming/recognition of app-created files within run
    directories."""

    @staticmethod
    def get_run_file_prefix(block_act_key: BlockActionKey) -> str:
        """
        Get the common prefix for files associated with a run.
        """
        return f"js_{block_act_key[0]}_block_{block_act_key[1]}_act_{block_act_key[2]}"

    @classmethod
    def get_commands_file_name(cls, block_act_key: BlockActionKey, shell: Shell) -> str:
        """
        Get the name of the file containing commands.
        """
        return cls.get_run_file_prefix(block_act_key) + shell.JS_EXT

    @classmethod
    def get_run_param_dump_file_prefix(cls, block_act_key: BlockActionKey) -> str:
        """Get the prefix to a file in the run directory that the app will dump parameter
        data to."""
        return cls.get_run_file_prefix(block_act_key) + "_inputs"

    @classmethod
    def get_run_param_load_file_prefix(cls, block_act_key: BlockActionKey) -> str:
        """Get the prefix to a file in the run directory that the app will load parameter
        data from."""
        return cls.get_run_file_prefix(block_act_key) + "_outputs"

    @classmethod
    def take_snapshot(cls, root_path=None) -> dict[str, Any]:
        """
        Take a :py:class:`JSONLikeDirSnapShot`.

        This includes command files that are invoked by jobscripts, the app log file, and
        the app standard out/error file.
        """
        snapshot = JSONLikeDirSnapShot()
        snapshot.take(root_path or ".")
        ss_js = snapshot.to_json_like(use_strings=True)
        ss_js.pop("root_path")  # always the current working directory of the run
        return ss_js



"""
Errors from the workflow system.
"""

from __future__ import annotations
import os
from collections.abc import Iterable, Mapping, Sequence
from textwrap import indent
from typing import Any, TYPE_CHECKING, Literal

from hpcflow.sdk.utils.strings import capitalise_first_letter

if TYPE_CHECKING:
    from logging import Logger
    from .enums import ParallelMode
    from .object_list import WorkflowLoopList
    from .parameters import InputSource, ValueSequence, SchemaInput
    from .types import ActionData
    from .task import WorkflowTask, Task


class InputValueDuplicateSequenceAddress(ValueError):
    """
    An InputValue has the same sequence address twice.
    """

    # FIXME: never used


class TaskTemplateMultipleSchemaObjectives(ValueError):
    """
    A TaskTemplate has multiple objectives.
    """

    def __init__(self, names: set[str]) -> None:
        super().__init__(
            f"All task schemas used within a task must have the same "
            f"objective, but found multiple objectives: {sorted(names)!r}"
        )


class TaskTemplateUnexpectedInput(ValueError):
    """
    A TaskTemplate was given unexpected input.
    """

    def __init__(self, unexpected_types: set[str]) -> None:
        super().__init__(
            f"The following input parameters are unexpected: {sorted(unexpected_types)!r}"
        )


class TaskTemplateUnexpectedSequenceInput(ValueError):
    """
    A TaskTemplate was given an unexpected sequence.
    """

    def __init__(
        self, inp_type: str, expected_types: set[str], seq: ValueSequence
    ) -> None:
        allowed_str = ", ".join(f'"{in_typ}"' for in_typ in expected_types)
        super().__init__(
            f"The input type {inp_type!r} specified in the following sequence"
            f" path is unexpected: {seq.path!r}. Available input types are: "
            f"{allowed_str}."
        )


class TaskTemplateMultipleInputValues(ValueError):
    """
    A TaskTemplate had multiple input values bound over each other.
    """

    def __init__(self, msg: str) -> None:
        super().__init__(msg)


class InvalidIdentifier(ValueError):
    """
    A bad identifier name was given.
    """

    def __init__(self, name: str) -> None:
        super().__init__(f"Invalid string for identifier: {name!r}")


class MissingInputs(Exception):
    """
    Inputs were missing.

    Parameters
    ----------
    missing_inputs:
        The missing inputs.
    """

    # TODO: add links to doc pages for common user-exceptions?

    def __init__(self, task: Task, missing_inputs: Iterable[str]) -> None:
        self.missing_inputs = tuple(missing_inputs)
        missing_str = ", ".join(map(repr, missing_inputs))
        super().__init__(
            f"Task {task.name}: the following inputs have no sources: {missing_str}."
        )


class UnrequiredInputSources(ValueError):
    """
    Input sources were provided that were not required.

    Parameters
    ----------
    unrequired_sources:
        The input sources that were not required.
    """

    def __init__(self, unrequired_sources: Iterable[str]) -> None:
        self.unrequired_sources = frozenset(unrequired_sources)
        message = (
            f"The following input sources are not required but have been specified: "
            f'{", ".join(map(repr, sorted(self.unrequired_sources)))}.'
        )
        if any((bad := src).startswith("inputs.") for src in self.unrequired_sources):
            # reminder about how to specify input sources:
            message += (
                f" Note that input source keys should not be specified with the "
                f"'inputs.' prefix. Did you mean to specify "
                f"{bad.removeprefix('inputs.')!r} instead of {bad!r}?"
            )
        super().__init__(message)


class ExtraInputs(Exception):
    """
    Extra inputs were provided.

    Parameters
    ----------
    extra_inputs:
        The extra inputs.
    """

    def __init__(self, extra_inputs: set[str]) -> None:
        self.extra_inputs = frozenset(extra_inputs)
        super().__init__(
            f"The following inputs are not required, but have been passed: "
            f'{", ".join(f"{typ!r}" for typ in extra_inputs)}.'
        )


class UnavailableInputSource(ValueError):
    """
    An input source was not available.
    """

    def __init__(
        self, source: InputSource, path: str, avail: Sequence[InputSource]
    ) -> None:
        super().__init__(
            f"The input source {source.to_string()!r} is not "
            f"available for input path {path!r}. Available "
            f"input sources are: {[src.to_string() for src in avail]}."
        )


class InapplicableInputSourceElementIters(ValueError):
    """
    An input source element iteration was inapplicable."""

    def __init__(self, source: InputSource, elem_iters_IDs: Sequence[int] | None) -> None:
        super().__init__(
            f"The specified `element_iters` for input source "
            f"{source.to_string()!r} are not all applicable. "
            f"Applicable element iteration IDs for this input source "
            f"are: {elem_iters_IDs!r}."
        )


class NoCoincidentInputSources(ValueError):
    """
    Could not line up input sources to make an actual valid execution.
    """

    def __init__(self, name: str, task_ref: int) -> None:
        super().__init__(
            f"Task {name!r}: input sources from task {task_ref!r} have "
            f"no coincident applicable element iterations. Consider setting "
            f"the element set (or task) argument "
            f"`allow_non_coincident_task_sources` to `True`, which will "
            f"allow for input sources from the same task to use different "
            f"(non-coinciding) subsets of element iterations from the "
            f"source task."
        )


class TaskTemplateInvalidNesting(ValueError):
    """
    Invalid nesting in a task template.
    """

    def __init__(self, key: str, value: float) -> None:
        super().__init__(
            f"`nesting_order` must be >=0 for all keys, but for key {key!r}, value "
            f"of {value!r} was specified."
        )


class TaskSchemaSpecValidationError(Exception):
    """
    A task schema failed to validate.
    """

    # FIXME: never used


class WorkflowSpecValidationError(Exception):
    """
    A workflow failed to validate.
    """

    # FIXME: never used


class InputSourceValidationError(Exception):
    """
    An input source failed to validate.
    """

    # FIXME: never used


class EnvironmentSpecValidationError(Exception):
    """
    An environment specification failed to validate.
    """

    # FIXME: never used


class ParameterSpecValidationError(Exception):
    """
    A parameter specification failed to validate.
    """

    # FIXME: never used


class FileSpecValidationError(Exception):
    """
    A file specification failed to validate.
    """

    # FIXME: never used


class DuplicateExecutableError(ValueError):
    """
    The same executable was present twice in an executable environment.
    """

    def __init__(self, duplicate_labels: list) -> None:
        super().__init__(
            f"Executables must have unique `label`s within each environment, but "
            f"found label(s) multiple times: {duplicate_labels!r}"
        )


class MissingCompatibleActionEnvironment(Exception):
    """
    Could not find a compatible action environment.
    """

    def __init__(self, msg: str) -> None:
        super().__init__(f"No compatible environment is specified for the {msg}.")


class MissingActionEnvironment(Exception):
    """
    Could not find an action environment.
    """

    # FIXME: never used


class ActionEnvironmentMissingNameError(Exception):
    """
    An action environment was missing its name.
    """

    def __init__(self, environment: Mapping[str, Any]) -> None:
        super().__init__(
            "The action-environment environment specification must include a string "
            "`name` key, or be specified as string that is that name. Provided "
            f"environment key was {environment!r}."
        )


class FromSpecMissingObjectError(Exception):
    """
    Missing object when deserialising from specification.
    """

    # FIXME: never used


class TaskSchemaMissingParameterError(Exception):
    """
    Parameter was missing from task schema.
    """

    # FIXME: never used


class ToJSONLikeChildReferenceError(Exception):
    """
    Failed to generate or reference a child object when converting to JSON.
    """

    # FIXME: never thrown


class InvalidInputSourceTaskReference(Exception):
    """
    Invalid input source in task reference.
    """

    def __init__(self, input_source: InputSource, task_ref: int | None = None) -> None:
        super().__init__(
            f"Input source {input_source.to_string()!r} cannot refer to the "
            f"outputs of its own task!"
            if task_ref is None
            else f"Input source {input_source.to_string()!r} refers to a missing "
            f"or inaccessible task: {task_ref!r}."
        )


class WorkflowNotFoundError(Exception):
    """
    Could not find the workflow.
    """

    def __init__(self, path, fs) -> None:
        super().__init__(
            f"Cannot infer a store format at path {path!r} with file system {fs!r}."
        )


class MalformedWorkflowError(Exception):
    """
    Workflow was a malformed document.
    """

    # TODO: use this class


class ValuesAlreadyPersistentError(Exception):
    """
    Trying to make a value persistent that already is so.
    """

    # FIXME: never used


class MalformedParameterPathError(ValueError):
    """
    The path to a parameter was ill-formed.
    """

    def __init__(self, msg: str) -> None:
        super().__init__(msg)


class MalformedNestingOrderPath(ValueError):
    """
    A nesting order path was ill-formed.
    """

    def __init__(self, k: str, allowed_nesting_paths: Iterable[str]) -> None:
        super().__init__(
            f"Element set: nesting order path {k!r} not understood. Each key in "
            f"`nesting_order` must be start with one of "
            f"{sorted(allowed_nesting_paths)!r}."
        )


class UnknownResourceSpecItemError(ValueError):
    """
    A resource specification item was not found.
    """

    def __init__(self, msg: str) -> None:
        super().__init__(msg)


class WorkflowParameterMissingError(AttributeError):
    """
    A parameter to a workflow was missing.
    """

    # FIXME: never thrown


class WorkflowBatchUpdateFailedError(Exception):
    """
    An update to a workflow failed.
    """

    # FIXME: only throw is commented out?


class WorkflowLimitsError(ValueError):
    """
    Workflow hit limits.
    """

    # FIXME: never used


class UnsetParameterDataErrorBase(Exception):
    """
    Exceptions related to attempts to retrieve unset parameters.
    """


class UnsetParameterDataError(UnsetParameterDataErrorBase):
    """
    Tried to read from an unset parameter.
    """

    def __init__(self, path: str | None, path_i: str) -> None:
        super().__init__(
            f"Element data path {path!r} resolves to unset data for "
            f"(at least) data-index path: {path_i!r}."
        )


class UnsetParameterFractionLimitExceededError(UnsetParameterDataErrorBase):
    """
    Given the specified `allow_failed_dependencies`, the fraction of failed dependencies
    (unset parameter data) is too high."""

    def __init__(
        self,
        schema_inp: SchemaInput,
        task: WorkflowTask,
        unset_fraction: float,
        log: Logger | None = None,
    ):
        msg = (
            f"Input {schema_inp.parameter.typ!r} of task {task.name!r}: higher "
            f"proportion of dependencies failed ({unset_fraction!r}) than allowed "
            f"({schema_inp.allow_failed_dependencies!r})."
        )
        if log:
            log.info(msg)
        super().__init__(msg)


class UnsetParameterNumberLimitExceededError(UnsetParameterDataErrorBase):
    """
    Given the specified `allow_failed_dependencies`, the number of failed dependencies
    (unset parameter data) is too high."""

    def __init__(
        self,
        schema_inp: SchemaInput,
        task: WorkflowTask,
        unset_num: int,
        log: Logger | None = None,
    ):
        msg = (
            f"Input {schema_inp.parameter.typ!r} of task {task.name!r}: higher number of "
            f"dependencies failed ({unset_num!r}) than allowed "
            f"({schema_inp.allow_failed_dependencies!r})."
        )
        if log:
            log.info(msg)
        super().__init__(msg)


class LoopAlreadyExistsError(Exception):
    """
    A particular loop (or its name) already exists.
    """

    def __init__(self, loop_name: str, loops: WorkflowLoopList) -> None:
        super().__init__(
            f"A loop with the name {loop_name!r} already exists in the workflow: "
            f"{getattr(loops, loop_name)!r}."
        )


class LoopTaskSubsetError(ValueError):
    """
    Problem constructing a subset of a task for a loop.
    """

    def __init__(self, loop_name: str, task_indices: Sequence[int]) -> None:
        super().__init__(
            f"Loop {loop_name!r}: task subset must be an ascending contiguous range, "
            f"but specified task indices were: {task_indices!r}."
        )


class SchedulerVersionsFailure(RuntimeError):
    """We couldn't get the scheduler and or shell versions."""

    # FIXME: unused

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)


class JobscriptSubmissionFailure(RuntimeError):
    """
    A job script could not be submitted to the scheduler.
    """

    def __init__(
        self,
        message: str,
        *,
        submit_cmd: list[str],
        js_idx: int,
        js_path: str,
        stdout: str | None = None,
        stderr: str | None = None,
        subprocess_exc: Exception | None = None,
        job_ID_parse_exc: Exception | None = None,
    ):
        self.message = message
        self.submit_cmd = submit_cmd
        self.js_idx = js_idx
        self.js_path = js_path
        self.stdout = stdout
        self.stderr = stderr
        self.subprocess_exc = subprocess_exc
        self.job_ID_parse_exc = job_ID_parse_exc
        super().__init__(message)


class SubmissionFailure(RuntimeError):
    """
    A job submission failed.
    """

    def __init__(
        self,
        sub_idx: int,
        submitted_js_idx: Sequence[int],
        exceptions: Iterable[JobscriptSubmissionFailure],
    ) -> None:
        msg = f"Some jobscripts in submission index {sub_idx} could not be submitted"
        if submitted_js_idx:
            msg += f" (but jobscripts {submitted_js_idx} were submitted successfully):"
        else:
            msg += ":"

        msg += "\n"
        for sub_err in exceptions:
            msg += (
                f"Jobscript {sub_err.js_idx} at path: {str(sub_err.js_path)!r}\n"
                f"Submit command: {sub_err.submit_cmd!r}.\n"
                f"Reason: {sub_err.message!r}\n"
            )
            if sub_err.subprocess_exc is not None:
                msg += f"Subprocess exception: {sub_err.subprocess_exc}\n"
            if sub_err.job_ID_parse_exc is not None:
                msg += f"Subprocess job ID parse exception: {sub_err.job_ID_parse_exc}\n"
            if sub_err.job_ID_parse_exc is not None:
                msg += f"Job ID parse exception: {sub_err.job_ID_parse_exc}\n"
            if sub_err.stdout:
                msg += f"Submission stdout:\n{indent(sub_err.stdout, '  ')}\n"
            if sub_err.stderr:
                msg += f"Submission stderr:\n{indent(sub_err.stderr, '  ')}\n"
        self.message = msg
        super().__init__(msg)


class WorkflowSubmissionFailure(RuntimeError):
    """
    A workflow submission failed.
    """

    def __init__(self, exceptions: Sequence[SubmissionFailure]) -> None:
        super().__init__("\n" + "\n\n".join(exn.message for exn in exceptions))


class ResourceValidationError(ValueError):
    """An incompatible resource requested by the user."""


class UnsupportedOSError(ResourceValidationError):
    """This machine is not of the requested OS."""

    def __init__(self, os_name: str) -> None:
        message = (
            f"OS {os_name!r} is not compatible with this machine/instance with OS: "
            f"{os.name!r}."
        )
        super().__init__(message)
        self.os_name = os_name


class UnsupportedShellError(ResourceValidationError):
    """We don't support this shell on this OS."""

    def __init__(self, shell: str, supported: Iterable[str]) -> None:
        sup = set(supported)
        message = (
            f"Shell {shell!r} is not supported on this machine/instance. Supported "
            f"shells are: {sup!r}."
        )
        super().__init__(message)
        self.shell = shell
        self.supported = frozenset(sup)


class UnsupportedSchedulerError(ResourceValidationError):
    """This scheduler is not supported on this machine according to the config.

    This is also raised in config validation when attempting to add a scheduler that is
    not known for this OS.

    """

    def __init__(
        self,
        scheduler: str,
        supported: Iterable[str] | None = None,
        available: Iterable[str] | None = None,
    ) -> None:
        if supported is not None:
            message = (
                f"Scheduler {scheduler!r} is not supported on this machine/instance. "
                f"Supported schedulers according to the app configuration are: "
                f"{supported!r}."
            )
        elif available is not None:
            message = (
                f"Scheduler {scheduler!r} is not supported on this OS. Schedulers "
                f"compatible with this OS are: {available!r}."
            )
        super().__init__(message)
        self.scheduler = scheduler
        self.supported = None if supported is None else tuple(supported)
        self.available = None if available is None else tuple(available)


class UnknownSGEPEError(ResourceValidationError):
    """
    Miscellaneous error from SGE parallel environment.
    """

    def __init__(self, env_name: str, all_env_names: Iterable[str]) -> None:
        super().__init__(
            f"The SGE parallel environment {env_name!r} is not "
            f"specified in the configuration. Specified parallel environments "
            f"are {sorted(all_env_names)!r}."
        )


class IncompatibleSGEPEError(ResourceValidationError):
    """
    The SGE parallel environment selected is incompatible.
    """

    def __init__(self, env_name: str, num_cores: int | None) -> None:
        super().__init__(
            f"The SGE parallel environment {env_name!r} is not "
            f"compatible with the number of cores requested: "
            f"{num_cores!r}."
        )


class NoCompatibleSGEPEError(ResourceValidationError):
    """
    No SGE parallel environment is compatible with request.
    """

    def __init__(self, num_cores: int | None) -> None:
        super().__init__(
            f"No compatible SGE parallel environment could be found for the "
            f"specified `num_cores` ({num_cores!r})."
        )


class IncompatibleParallelModeError(ResourceValidationError):
    """
    The parallel mode is incompatible.
    """

    def __init__(self, parallel_mode: ParallelMode) -> None:
        super().__init__(
            f"For the {parallel_mode.name.lower()} parallel mode, "
            f"only a single node may be requested."
        )


class UnknownSLURMPartitionError(ResourceValidationError):
    """
    The requested SLURM partition isn't known.
    """

    def __init__(self, part_name: str, all_parts: Iterable[str]) -> None:
        super().__init__(
            f"The SLURM partition {part_name!r} is not "
            f"specified in the configuration. Specified partitions are "
            f"{sorted(all_parts)!r}."
        )


class IncompatibleSLURMPartitionError(ResourceValidationError):
    """
    The requested SLURM partition is incompatible.
    """

    def __init__(self, part_name: str, attr_kind: str, value) -> None:
        super().__init__(
            f"The SLURM partition {part_name!r} is not "
            f"compatible with the {attr_kind} requested: {value!r}."
        )


class IncompatibleSLURMArgumentsError(ResourceValidationError):
    """
    The SLURM arguments are incompatible with each other.
    """

    def __init__(self, msg: str) -> None:
        super().__init__(msg)


class _MissingStoreItemError(ValueError):
    def __init__(self, id_lst: Iterable[int], item_type: str) -> None:
        message = (
            f"Store {item_type}s with the following IDs do not all exist: {id_lst!r}"
        )
        super().__init__(message)
        self.id_lst = id_lst


class MissingStoreTaskError(_MissingStoreItemError):
    """Some task IDs do not exist."""

    _item_type = "task"

    def __init__(self, id_lst: Iterable[int]) -> None:
        super().__init__(id_lst, self._item_type)


class MissingStoreElementError(_MissingStoreItemError):
    """Some element IDs do not exist."""

    _item_type = "element"

    def __init__(self, id_lst: Iterable[int]) -> None:
        super().__init__(id_lst, self._item_type)


class MissingStoreElementIterationError(_MissingStoreItemError):
    """Some element iteration IDs do not exist."""

    _item_type = "element iteration"

    def __init__(self, id_lst: Iterable[int]) -> None:
        super().__init__(id_lst, self._item_type)


class MissingStoreEARError(_MissingStoreItemError):
    """Some EAR IDs do not exist."""

    _item_type = "EAR"

    def __init__(self, id_lst: Iterable[int]) -> None:
        super().__init__(id_lst, self._item_type)


class MissingParameterData(_MissingStoreItemError):
    """Some parameter IDs do not exist"""

    _item_type = "parameter"

    def __init__(self, id_lst: Iterable[int]) -> None:
        super().__init__(id_lst, self._item_type)


class ParametersMetadataReadOnlyError(RuntimeError):
    pass


class NotSubmitMachineError(RuntimeError):
    """
    The requested machine can't be submitted to.
    """

    def __init__(self) -> None:
        super().__init__(
            "Cannot get active state of the jobscript because the current machine "
            "is not the machine on which the jobscript was submitted."
        )


class RunNotAbortableError(ValueError):
    """
    Cannot abort the run.
    """

    def __init__(self) -> None:
        super().__init__(
            "The run is not defined as abortable in the task schema, so it cannot "
            "be aborted."
        )


class NoCLIFormatMethodError(AttributeError):
    """
    Some CLI class lacks a format method
    """

    def __init__(self, method: str, inp_val: object) -> None:
        super().__init__(
            f"No CLI format method {method!r} exists for the object {inp_val!r}."
        )


class ContainerKeyError(KeyError):
    """
    A key could not be mapped in a container.

    Parameters
    ----------
    path:
        The path whose resolution failed.
    """

    def __init__(self, path: list[str]) -> None:
        self.path = path
        super().__init__()


class MayNeedObjectError(Exception):
    """
    An object is needed but not present.

    Parameters
    ----------
    path:
        The path whose resolution failed.
    """

    def __init__(self, path: str) -> None:
        self.path = path
        super().__init__()


class NoAvailableElementSetsError(Exception):
    """
    No element set is available.
    """

    def __init__(self) -> None:
        super().__init__()


class OutputFileParserNoOutputError(ValueError):
    """
    There was no output for the output file parser to parse.
    """

    def __init__(self) -> None:
        super().__init__()


class SubmissionEnvironmentError(ValueError):
    """
    Raised when submitting a workflow on a machine without a compatible environment.
    """


def _spec_to_ref(env_spec: Mapping[str, Any]):
    non_name_spec = {k: v for k, v in env_spec.items() if k != "name"}
    spec_str = f" with specifiers {non_name_spec!r}" if non_name_spec else ""
    return f"{env_spec['name']!r}{spec_str}"


class MissingEnvironmentExecutableError(SubmissionEnvironmentError):
    """
    The environment does not have the requested executable at all.
    """

    def __init__(self, env_spec: Mapping[str, Any], exec_label: str) -> None:
        super().__init__(
            f"The environment {_spec_to_ref(env_spec)} as defined on this machine has no "
            f"executable labelled {exec_label!r}, which is required for this "
            f"submission, so the submission cannot be created."
        )


class MissingEnvironmentExecutableInstanceError(SubmissionEnvironmentError):
    """
    The environment does not have a suitable instance of the requested executable.
    """

    def __init__(
        self, env_spec: Mapping[str, Any], exec_label: str, js_idx: int, res: dict
    ) -> None:
        super().__init__(
            f"No matching executable instances found for executable "
            f"{exec_label!r} of environment {_spec_to_ref(env_spec)} for jobscript "
            f"index {js_idx!r} with requested resources {res!r}."
        )


class MissingEnvironmentError(SubmissionEnvironmentError):
    """
    There is no environment with that name.
    """

    def __init__(self, env_spec: Mapping[str, Any]) -> None:
        super().__init__(
            f"The environment {_spec_to_ref(env_spec)} is not defined on this machine, so the "
            f"submission cannot be created."
        )


class UnsupportedActionDataFormat(ValueError):
    """
    That format of script or program data is not supported.
    """

    def __init__(
        self,
        type: Literal["script", "program"],
        data: ActionData,
        kind: Literal["input", "output"],
        name: str,
        formats: tuple[str, ...],
    ) -> None:
        super().__init__(
            f"{capitalise_first_letter(type)} data format {data!r} for {kind} parameter "
            f"{name!r} is not understood. Available {type} data formats are: "
            f"{formats!r}."
        )


class UnknownActionDataParameter(ValueError):
    """
    Unknown parameter in script or program data.
    """

    def __init__(
        self,
        type: Literal["script", "program"],
        name: str,
        direction: Literal["in", "out"],
        param_names: Sequence[str],
    ) -> None:
        super().__init__(
            f"{capitalise_first_letter(type)} data {direction} parameter {name!r} is not "
            f"a known parameter of the action. Parameters are: {param_names!r}."
        )


class UnknownActionDataKey(ValueError):
    """
    Unknown key in script data.
    """

    def __init__(
        self, type: Literal["script", "program"], key: str, allowed_keys: Sequence[str]
    ) -> None:
        super().__init__(
            f"{capitalise_first_letter(type)} data key {key!r} is not understood. "
            f"Allowed keys are: {allowed_keys!r}."
        )


class MissingVariableSubstitutionError(KeyError):
    """
    No definition available of a variable being substituted.
    """

    def __init__(self, var_name: str, variables: Iterable[str]) -> None:
        super().__init__(
            f"The variable {var_name!r} referenced in the string does not match "
            f"any of the provided variables: {sorted(variables)!r}."
        )


class EnvironmentPresetUnknownEnvironmentError(ValueError):
    """
    An environment preset could not be resolved to an execution environment.
    """

    def __init__(self, name: str, bad_envs: Iterable[str]) -> None:
        super().__init__(
            f"Task schema {name} has environment presets that refer to one "
            f"or more environments that are not referenced in any of the task "
            f"schema's actions: {', '.join(f'{env!r}' for env in sorted(bad_envs))}."
        )


class UnknownEnvironmentPresetError(ValueError):
    """
    An execution environment was unknown.
    """

    def __init__(self, preset_name: str, schema_name: str) -> None:
        super().__init__(
            f"There is no environment preset named {preset_name!r} defined "
            f"in the task schema {schema_name}."
        )


class MultipleEnvironmentsError(ValueError):
    """
    Multiple applicable execution environments exist.
    """

    def __init__(self, env_spec: Mapping[str, Any]) -> None:
        super().__init__(
            f"Multiple environments {_spec_to_ref(env_spec)} are defined on this machine."
        )


class MissingElementGroup(ValueError):
    """
    An element group should exist but doesn't.
    """

    def __init__(self, task_name: str, group_name: str, input_path: str) -> None:
        super().__init__(
            f"Adding elements to task {task_name!r}: "
            f"no element group named {group_name!r} found for input {input_path!r}."
        )


class YAMLError(ValueError):
    """
    A problem with parsing a YAML file.
    """



import json
import os
from pathlib import Path
import shutil
import time
import pytest

from hpcflow.app import app as hf
from hpcflow.sdk.core.enums import EARStatus
from hpcflow.sdk.core.test_utils import P1_parameter_cls as P1

# note: when testing the frozen app, we might not have MatFlow installed in the built in
# python_env MatFlow environment, so we should skip these tests.


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
@pytest.mark.parametrize("combine_scripts", [False, True])
def test_script_direct_in_direct_out(null_config, tmp_path: Path, combine_scripts: bool):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_direct_in_direct_out.py>>",
                script_data_in="direct",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    p1_val = 101
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="main_script_test",
        path=tmp_path,
        resources={"any": {"combine_scripts": combine_scripts}},
    )
    wk.submit(wait=True, add_to_known=False)

    p2 = wk.tasks[0].elements[0].outputs.p2
    assert isinstance(p2, hf.ElementParameter)
    assert p2.value == p1_val + 100


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
@pytest.mark.parametrize("combine_scripts", [False, True])
def test_script_direct_sub_param_in_direct_out(
    null_config, tmp_path: Path, combine_scripts: bool
):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_direct_sub_param_in_direct_out.py>>",
                script_data_in={"p1.a": "direct"},
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    p1_val = {"a": 101}
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="main_script_test",
        path=tmp_path,
        resources={"any": {"combine_scripts": combine_scripts}},
    )
    wk.submit(wait=True, add_to_known=False)

    p2 = wk.tasks[0].elements[0].outputs.p2
    assert isinstance(p2, hf.ElementParameter)
    assert p2.value == p1_val["a"] + 100


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
@pytest.mark.parametrize("combine_scripts", [False, True])
def test_script_direct_in_direct_out_single_label(
    null_config, tmp_path: Path, combine_scripts: bool
):
    """This uses the same test script as the `test_script_direct_in_direct_out` test;
    single labels are trivial and need not be referenced in the script."""
    p1_label = "one"
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"), labels={p1_label: {}})],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_direct_in_direct_out.py>>",
                script_data_in="direct",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    p1_val = 101
    t1 = hf.Task(schema=s1, inputs={f"p1[{p1_label}]": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="main_script_test",
        path=tmp_path,
        resources={"any": {"combine_scripts": combine_scripts}},
    )
    wk.submit(wait=True, add_to_known=False)

    p2 = wk.tasks[0].elements[0].outputs.p2
    assert isinstance(p2, hf.ElementParameter)
    assert p2.value == p1_val + 100


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
@pytest.mark.parametrize("combine_scripts", [False, True])
def test_script_direct_in_direct_out_labels(
    null_config, tmp_path: Path, combine_scripts: bool
):
    p1_label_1 = "one"
    p1_label_2 = "two"
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[
            hf.SchemaInput(
                parameter=hf.Parameter("p1"),
                labels={p1_label_1: {}, p1_label_2: {}},
                multiple=True,
            )
        ],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_direct_in_direct_out_labels.py>>",
                script_data_in="direct",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    p1_1_val = 101
    p1_2_val = 201
    t1 = hf.Task(
        schema=s1,
        inputs={
            f"p1[{p1_label_1}]": p1_1_val,
            f"p1[{p1_label_2}]": p1_2_val,
        },
    )
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="main_script_test",
        path=tmp_path,
        resources={"any": {"combine_scripts": combine_scripts}},
    )
    wk.submit(wait=True, add_to_known=False)

    p2 = wk.tasks[0].elements[0].outputs.p2
    assert isinstance(p2, hf.ElementParameter)
    assert p2.value == p1_1_val + p1_2_val


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
@pytest.mark.parametrize("combine_scripts", [False, True])
def test_script_json_in_json_out(null_config, tmp_path: Path, combine_scripts: bool):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_json_in_json_out.py>>",
                script_data_in="json",
                script_data_out="json",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
                requires_dir=True,
            )
        ],
    )
    p1_val = 101
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="main_script_test",
        path=tmp_path,
        resources={"any": {"combine_scripts": combine_scripts}},
    )
    wk.submit(wait=True, add_to_known=False)

    p2 = wk.tasks[0].elements[0].outputs.p2
    assert isinstance(p2, hf.ElementParameter)
    assert p2.value == p1_val + 100


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
@pytest.mark.parametrize("combine_scripts", [False, True])
def test_script_json_in_json_out_labels(
    null_config, tmp_path: Path, combine_scripts: bool
):
    p1_label_1 = "one"
    p1_label_2 = "two"
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[
            hf.SchemaInput(
                parameter=hf.Parameter("p1"),
                labels={p1_label_1: {}, p1_label_2: {}},
                multiple=True,
            )
        ],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_json_in_json_out_labels.py>>",
                script_data_in="json",
                script_data_out="json",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
                requires_dir=True,
            )
        ],
    )
    p1_1_val = 101
    p1_2_val = 201
    t1 = hf.Task(
        schema=s1,
        inputs={
            f"p1[{p1_label_1}]": p1_1_val,
            f"p1[{p1_label_2}]": p1_2_val,
        },
    )
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="main_script_test",
        path=tmp_path,
        resources={"any": {"combine_scripts": combine_scripts}},
    )
    wk.submit(wait=True, add_to_known=False)

    p2 = wk.tasks[0].elements[0].outputs.p2
    assert isinstance(p2, hf.ElementParameter)
    assert p2.value == p1_1_val + p1_2_val


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
@pytest.mark.parametrize("combine_scripts", [False, True])
def test_script_json_sub_param_in_json_out_labels(
    null_config, tmp_path: Path, combine_scripts: bool
):
    p1_label_1 = "one"
    p1_label_2 = "two"
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[
            hf.SchemaInput(
                parameter=hf.Parameter("p1"),
                labels={p1_label_1: {}, p1_label_2: {}},
                multiple=True,
            )
        ],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_json_sub_param_in_json_out_labels.py>>",
                script_data_in={"p1[one].a": "json", "p1[two]": "json"},
                script_data_out="json",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
                requires_dir=True,
            )
        ],
    )
    a_val = 101
    p1_2_val = 201
    t1 = hf.Task(
        schema=s1,
        inputs={
            f"p1[{p1_label_1}]": {"a": a_val},
            f"p1[{p1_label_2}]": p1_2_val,
        },
    )
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="main_script_test",
        path=tmp_path,
        resources={"any": {"combine_scripts": combine_scripts}},
    )
    wk.submit(wait=True, add_to_known=False)

    p2 = wk.tasks[0].elements[0].outputs.p2
    assert isinstance(p2, hf.ElementParameter)
    assert p2.value == a_val + p1_2_val


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
@pytest.mark.parametrize("combine_scripts", [False, True])
def test_script_json_and_direct_in_json_out(
    null_config, tmp_path: Path, combine_scripts: bool
):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[
            hf.SchemaInput(parameter=hf.Parameter("p1")),
            hf.SchemaInput(parameter=hf.Parameter("p2")),
        ],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p3"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_json_and_direct_in_json_out.py>>",
                script_data_in={"p1": "json", "p2": "direct"},
                script_data_out="json",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
                requires_dir=True,
            )
        ],
    )
    p1_val = 101
    p2_val = 201
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val, "p2": p2_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="main_script_test",
        path=tmp_path,
        resources={"any": {"combine_scripts": combine_scripts}},
    )
    wk.submit(wait=True, add_to_known=False)

    p3 = wk.tasks[0].elements[0].outputs.p3
    assert isinstance(p3, hf.ElementParameter)
    assert p3.value == p1_val + p2_val


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
@pytest.mark.parametrize("combine_scripts", [False, True])
def test_script_json_in_json_and_direct_out(
    null_config, tmp_path: Path, combine_scripts: bool
):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[
            hf.SchemaInput(parameter=hf.Parameter("p2")),
            hf.SchemaOutput(parameter=hf.Parameter("p3")),
        ],
        actions=[
            hf.Action(
                script="<<script:main_script_test_json_in_json_and_direct_out.py>>",
                script_data_in="json",
                script_data_out={"p2": "json", "p3": "direct"},
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
                requires_dir=True,
            )
        ],
    )
    p1_val = 101
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="main_script_test",
        path=tmp_path,
        resources={"any": {"combine_scripts": combine_scripts}},
    )
    wk.submit(wait=True, add_to_known=False)

    p2 = wk.tasks[0].elements[0].outputs.p2
    assert isinstance(p2, hf.ElementParameter)
    p3 = wk.tasks[0].elements[0].outputs.p3
    assert isinstance(p3, hf.ElementParameter)
    assert p2.value == p1_val + 100
    assert p3.value == p1_val + 200


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
@pytest.mark.parametrize("combine_scripts", [False, True])
def test_script_json_in_obj(null_config, tmp_path: Path, combine_scripts: bool):
    """Use a custom JSON dumper defined in the P1 class."""
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1c"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_json_in_obj.py>>",
                script_data_in="json",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
                requires_dir=True,
            )
        ],
        parameter_class_modules=["hpcflow.sdk.core.test_utils"],
    )
    a_val = 1
    t1 = hf.Task(schema=s1, inputs={"p1c": P1(a=a_val)})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="main_script_test",
        path=tmp_path,
        resources={"any": {"combine_scripts": combine_scripts}},
    )
    wk.submit(wait=True, add_to_known=False)

    p2 = wk.tasks[0].elements[0].outputs.p2
    assert isinstance(p2, hf.ElementParameter)
    assert p2.value == a_val + 100


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
@pytest.mark.parametrize("combine_scripts", [False, True])
def test_script_hdf5_in_obj(null_config, tmp_path: Path, combine_scripts: bool):
    """Use a custom HDF5 dumper defined in the P1 class."""
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1c"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_hdf5_in_obj.py>>",
                script_data_in="hdf5",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
                requires_dir=True,
            )
        ],
        parameter_class_modules=["hpcflow.sdk.core.test_utils"],
    )
    a_val = 1
    t1 = hf.Task(schema=s1, inputs={"p1c": P1(a=a_val)})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="main_script_test",
        path=tmp_path,
        resources={"any": {"combine_scripts": combine_scripts}},
    )
    wk.submit(wait=True, add_to_known=False)

    p2 = wk.tasks[0].elements[0].outputs.p2
    assert isinstance(p2, hf.ElementParameter)
    assert p2.value == a_val + 100


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
@pytest.mark.parametrize("combine_scripts", [False, True])
def test_script_hdf5_in_obj_group(null_config, tmp_path: Path, combine_scripts: bool):
    s0 = hf.TaskSchema(
        objective="define_p1c",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1c"))],
    )
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1c"), group="my_group")],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_hdf5_in_obj_group.py>>",
                script_data_in="hdf5",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
                requires_dir=True,
            )
        ],
        parameter_class_modules=["hpcflow.sdk.core.test_utils"],
    )
    a_vals = (1, 2)
    t0 = hf.Task(
        schema=s0,
        sequences=[hf.ValueSequence(path="inputs.p1c", values=[P1(a=i) for i in a_vals])],
        groups=[hf.ElementGroup("my_group")],
    )
    t1 = hf.Task(schema=s1)
    wk = hf.Workflow.from_template_data(
        tasks=[t0, t1],
        template_name="main_script_test",
        path=tmp_path,
        resources={"any": {"combine_scripts": combine_scripts}},
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    p2 = wk.tasks[1].elements[0].outputs.p2
    assert isinstance(p2, hf.ElementParameter)
    assert p2.value == sum(a_vals) + 100


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
@pytest.mark.parametrize("combine_scripts", [False, True])
def test_script_json_out_obj(null_config, tmp_path: Path, combine_scripts: bool):
    """Use a custom JSON saver defined in the P1 class."""
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p1c"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_json_out_obj.py>>",
                script_data_in="direct",
                script_data_out="json",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
                requires_dir=True,
            )
        ],
        parameter_class_modules=["hpcflow.sdk.core.test_utils"],
    )
    p1_val = 1
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="main_script_test",
        path=tmp_path,
        resources={"any": {"combine_scripts": combine_scripts}},
    )
    wk.submit(wait=True, add_to_known=False)

    p1c = wk.tasks[0].elements[0].outputs.p1c
    assert isinstance(p1c, hf.ElementParameter)
    assert p1c.value == P1(a=p1_val + 100)


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
@pytest.mark.parametrize("combine_scripts", [False, True])
def test_script_hdf5_out_obj(null_config, tmp_path: Path, combine_scripts: bool):
    """Use a custom HDF5 saver defined in the P1 class."""
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p1c"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_hdf5_out_obj.py>>",
                script_data_in="direct",
                script_data_out="hdf5",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
                requires_dir=True,
            )
        ],
        parameter_class_modules=["hpcflow.sdk.core.test_utils"],
    )
    p1_val = 1
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="main_script_test",
        path=tmp_path,
        resources={"any": {"combine_scripts": combine_scripts}},
    )
    wk.submit(wait=True, add_to_known=False)

    p1c = wk.tasks[0].elements[0].outputs.p1c
    assert isinstance(p1c, hf.ElementParameter)
    assert p1c.value == P1(a=p1_val + 100)


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
@pytest.mark.parametrize("combine_scripts", [False, True])
def test_script_direct_in_pass_env_spec(
    new_null_config, tmp_path: Path, combine_scripts: bool
):

    vers_spec = {"version": "1.2"}
    env = hf.Environment(
        name="python_env_with_specifiers",
        specifiers=vers_spec,
        executables=[
            hf.Executable(
                label="python_script",
                instances=[
                    hf.ExecutableInstance(
                        command="python <<script_path>> <<args>>",
                        num_cores=1,
                        parallel_mode=None,
                    )
                ],
            )
        ],
    )
    hf.envs.add_object(env, skip_duplicates=True)

    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_direct_in_direct_out_env_spec.py>>",
                script_data_in="direct",
                script_data_out="direct",
                script_exe="python_script",
                script_pass_env_spec=True,
                environments=[
                    hf.ActionEnvironment(environment="python_env_with_specifiers")
                ],
            )
        ],
    )
    t1 = hf.Task(
        schema=s1,
        inputs={"p1": 101},
        environments={"python_env_with_specifiers": vers_spec},
    )
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="main_script_test",
        path=tmp_path,
        resources={"any": {"combine_scripts": combine_scripts}},
    )
    wk.submit(wait=True, add_to_known=False)

    p2 = wk.tasks[0].elements[0].outputs.p2
    assert isinstance(p2, hf.ElementParameter)
    assert p2.value == {
        "name": "python_env_with_specifiers",
        **vers_spec,
    }
    hf.reload_template_components()  # remove extra envs


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
@pytest.mark.parametrize("combine_scripts", [False, True])
def test_script_std_stream_redirect_on_exception(
    new_null_config, tmp_path: Path, combine_scripts: bool
):
    """Test exceptions raised by the app during execution of a script are printed to the
    std-stream redirect file (and not the jobscript's standard error file)."""

    # define a custom python environment which redefines the `WK_PATH` shell variable to
    # a nonsense value so the app cannot load the workflow and thus raises an exception
    app_caps = hf.package_name.upper()
    if os.name == "nt":
        env_cmd = f'$env:{app_caps}_WK_PATH = "nonsense_path"'
    else:
        env_cmd = f'export {app_caps}_WK_PATH="nonsense_path"'

    env_cmd += "; python <<script_path>> <<args>>"
    bad_env = hf.Environment(
        name="bad_python_env",
        executables=[
            hf.Executable(
                label="python_script",
                instances=[
                    hf.ExecutableInstance(
                        command=env_cmd,
                        num_cores=1,
                        parallel_mode=None,
                    )
                ],
            )
        ],
    )
    hf.envs.add_object(bad_env, skip_duplicates=True)

    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_direct_in_direct_out.py>>",
                script_data_in="direct",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="bad_python_env")],
            )
        ],
    )
    p1_val = 101
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="main_script_test",
        path=tmp_path,
        resources={"any": {"combine_scripts": combine_scripts}},
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    # jobscript stderr should be empty
    assert not wk.submissions[0].jobscripts[0].direct_stderr_path.read_text()

    # std stream file has workflow not found traceback
    if combine_scripts:
        std_stream_path = wk.submissions[0].jobscripts[0].get_app_std_path()
    else:
        run = wk.get_all_EARs()[0]
        std_stream_path = run.get_app_std_path()
    assert std_stream_path.is_file()
    assert "WorkflowNotFoundError" in std_stream_path.read_text()

    hf.reload_template_components()  # remove extra envs


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
@pytest.mark.parametrize("combine_scripts", [False, True])
def test_script_std_out_std_err_not_redirected(
    null_config, tmp_path: Path, combine_scripts: bool
):
    """Test that standard error and output streams from a script are written to the jobscript
    standard error and output files."""
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[
            hf.SchemaInput(parameter=hf.Parameter("stdout_msg")),
            hf.SchemaInput(parameter=hf.Parameter("stderr_msg")),
        ],
        actions=[
            hf.Action(
                script="<<script:main_script_test_std_out_std_err.py>>",
                script_data_in="direct",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    stdout_msg = "hello stdout!"
    stderr_msg = "hello stderr!"
    t1 = hf.Task(schema=s1, inputs={"stdout_msg": stdout_msg, "stderr_msg": stderr_msg})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="main_script_test",
        path=tmp_path,
        resources={"any": {"combine_scripts": combine_scripts}},
    )
    wk.submit(wait=True, add_to_known=False)

    if wk.submissions[0].jobscripts[0].resources.combine_jobscript_std:
        std_out_err = wk.submissions[0].jobscripts[0].direct_std_out_err_path.read_text()
        assert std_out_err.strip() == f"{stdout_msg}\n{stderr_msg}"
    else:
        std_out = wk.submissions[0].jobscripts[0].direct_stdout_path.read_text()
        std_err = wk.submissions[0].jobscripts[0].direct_stderr_path.read_text()
        assert std_out.strip() == stdout_msg
        assert std_err.strip() == stderr_msg


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
@pytest.mark.parametrize("combine_scripts", [False, True])
def test_script_pass_env_spec(null_config, tmp_path: Path, combine_scripts: bool):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:env_specifier_test/main_script_test_pass_env_spec.py>>",
                script_data_in="direct",
                script_data_out="direct",
                script_exe="python_script",
                script_pass_env_spec=True,
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    p1_val = 101
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="main_script_test_pass_env_spec",
        path=tmp_path,
        resources={"any": {"combine_scripts": combine_scripts}},
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    std_out = wk.submissions[0].jobscripts[0].direct_stdout_path.read_text().strip()
    p2 = wk.tasks[0].elements[0].outputs.p2
    assert isinstance(p2, hf.ElementParameter)
    assert p2.value == p1_val + 100
    assert std_out == "{'name': 'python_env'}"


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
@pytest.mark.parametrize("combine_scripts", [False, True])
def test_env_specifier_in_main_script_path(
    new_null_config, tmp_path: Path, combine_scripts: bool
):
    py_env = hf.Environment(
        name="python_env",
        specifiers={"version": "v1"},
        executables=[
            hf.Executable(
                label="python_script",
                instances=[
                    hf.ExecutableInstance(
                        command="python <<script_path>> <<args>>",
                        num_cores=1,
                        parallel_mode=None,
                    )
                ],
            )
        ],
    )
    hf.envs.add_object(py_env, skip_duplicates=True)

    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:env_specifier_test/<<env:version>>/main_script_test_direct_in_direct_out.py>>",
                script_data_in="direct",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )

    p1_val = 101
    t1 = hf.Task(
        schema=s1,
        inputs={"p1": p1_val},
        environments={"python_env": {"version": "v1"}},
    )
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="main_script_test_env_spec_script_path",
        path=tmp_path,
        resources={"any": {"combine_scripts": combine_scripts}},
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    p2 = wk.tasks[0].elements[0].outputs.p2
    assert isinstance(p2, hf.ElementParameter)
    assert p2.value == p1_val + 100

    hf.reload_template_components()  # remove extra envs


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
@pytest.mark.parametrize("combine_scripts", [False, True])
def test_env_specifier_in_main_script_path_multiple_scripts(
    new_null_config, tmp_path: Path, combine_scripts: bool
):
    """Test two elements with different environment specifiers use two distinct scripts"""
    py_env_v1 = hf.Environment(
        name="python_env",
        specifiers={"version": "v1"},
        executables=[
            hf.Executable(
                label="python_script",
                instances=[
                    hf.ExecutableInstance(
                        command="python <<script_path>> <<args>>",
                        num_cores=1,
                        parallel_mode=None,
                    )
                ],
            )
        ],
    )
    py_env_v2 = hf.Environment(
        name="python_env",
        specifiers={"version": "v2"},
        executables=[
            hf.Executable(
                label="python_script",
                instances=[
                    hf.ExecutableInstance(
                        command="python <<script_path>> <<args>>",
                        num_cores=1,
                        parallel_mode=None,
                    )
                ],
            )
        ],
    )
    hf.envs.add_objects([py_env_v1, py_env_v2], skip_duplicates=True)

    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:env_specifier_test/<<env:version>>/main_script_test_direct_in_direct_out.py>>",
                script_data_in="direct",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )

    p1_val = 101
    t1 = hf.Task(
        schema=s1,
        inputs={"p1": p1_val},
        environments={"python_env": {"version": "v1"}},
        sequences=[
            hf.ValueSequence(
                path="environments.python_env.version",
                values=["v1", "v2"],
            )
        ],
    )
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="main_script_test_multiple_env_spec_script",
        path=tmp_path,
        resources={"any": {"combine_scripts": combine_scripts}},
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    # v1 and v2 scripts output different values:
    e1, e2 = wk.tasks.t1.elements
    e1_p2 = e1.outputs.p2
    e2_p2 = e2.outputs.p2
    assert isinstance(e1_p2, hf.ElementParameter)
    assert isinstance(e2_p2, hf.ElementParameter)
    assert e1_p2.value == 201
    assert e2_p2.value == 301

    hf.reload_template_components()  # remove extra envs


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
@pytest.mark.parametrize("combine_scripts", [False, True])
def test_script_direct_in_direct_out_multi_element(
    null_config, tmp_path: Path, combine_scripts: bool
):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_direct_in_direct_out.py>>",
                script_data_in="direct",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    p1_vals = (101, 102, 103)
    t1 = hf.Task(
        schema=s1, sequences=[hf.ValueSequence(path="inputs.p1", values=p1_vals)]
    )
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="main_script_test_multi_element",
        path=tmp_path,
        resources={"any": {"combine_scripts": combine_scripts}},
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    e0_p2 = wk.tasks[0].elements[0].outputs.p2
    e1_p2 = wk.tasks[0].elements[1].outputs.p2
    e2_p2 = wk.tasks[0].elements[2].outputs.p2

    assert isinstance(e0_p2, hf.ElementParameter)
    assert isinstance(e1_p2, hf.ElementParameter)
    assert isinstance(e2_p2, hf.ElementParameter)

    assert e0_p2.value == p1_vals[0] + 100
    assert e1_p2.value == p1_vals[1] + 100
    assert e2_p2.value == p1_vals[2] + 100

    # check only one script generated, and its name:
    script_name, _ = t1.schema.actions[0].get_script_artifact_name(env_spec={}, act_idx=0)
    script_files = list(i.name for i in wk.submissions[0].scripts_path.glob("*"))
    assert len(script_files) == 1
    assert script_files[0] == script_name if not combine_scripts else "js_0.py"


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_repeated_action_in_schema(null_config, tmp_path: Path):
    # TODO: cannot currently use same Action object multiple times in a schema
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_direct_in_direct_out.py>>",
                script_data_in="direct",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            ),
            hf.Action(
                script="<<script:main_script_test_direct_in_direct_out.py>>",
                script_data_in="direct",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            ),
        ],
    )
    p1_val = 101
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="test_repeated_action_in_schema",
        path=tmp_path,
        resources={"any": {"write_app_logs": True}},
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    # check scripts generated for act 0 and 1 have the same contents
    act_0_script, _ = wk.tasks.t1.template.schema.actions[0].get_script_artifact_name(
        env_spec={}, act_idx=0
    )
    act_1_script, _ = wk.tasks.t1.template.schema.actions[1].get_script_artifact_name(
        env_spec={}, act_idx=1
    )
    act_0_script_path = wk.submissions[0].scripts_path / act_0_script
    act_1_script_path = wk.submissions[0].scripts_path / act_1_script
    assert act_0_script_path.read_text() == act_1_script_path.read_text()

    # the two files will be symlinked if not on Windows (may be symlinked on Windows,
    # depending on if user is admin)
    if os.name != "nt":
        assert act_1_script_path.is_symlink()

    # output will be taken from second action
    p2 = wk.tasks[0].elements[0].outputs.p2
    assert isinstance(p2, hf.ElementParameter)
    assert p2.value == p1_val + 100


# TODO: same action with different env spec path (v1/v2) in same schema (check contents
# different!). Cannot yet do this because it is not possible to set environment spec
# for diferrent "main" actions within the same task.


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_main_script_two_schemas_same_action(null_config, tmp_path: Path):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_direct_in_direct_out.py>>",
                script_data_in="direct",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            ),
        ],
    )
    s2 = hf.TaskSchema(
        objective="t2",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_direct_in_direct_out.py>>",
                script_data_in="direct",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            ),
        ],
    )
    p1_val = 101
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    t2 = hf.Task(schema=s2, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1, t2],
        template_name="main_script_test_two_schemas_same_action",
        path=tmp_path,
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    # check scripts generated for t1 and t2 have the same contents
    t1_script, _ = wk.tasks.t1.template.schema.actions[0].get_script_artifact_name(
        env_spec={}, act_idx=0
    )
    t2_script, _ = wk.tasks.t2.template.schema.actions[0].get_script_artifact_name(
        env_spec={}, act_idx=0
    )
    t1_script_path = wk.submissions[0].scripts_path / t1_script
    t2_script_path = wk.submissions[0].scripts_path / t2_script
    assert t1_script_path.read_text() == t2_script_path.read_text()

    # the two files will be symlinked if not on Windows (may be symlinked on Windows,
    # depending on if user is admin)
    if os.name != "nt":
        assert t2_script_path.is_symlink()

    # check output
    t0_p2 = wk.tasks[0].elements[0].outputs.p2
    t1_p2 = wk.tasks[1].elements[0].outputs.p2
    assert isinstance(t0_p2, hf.ElementParameter)
    assert isinstance(t1_p2, hf.ElementParameter)
    assert t0_p2.value == p1_val + 100
    assert t1_p2.value == p1_val + 100

    # now copy the workflow elsewhere and check the symlink between the scripts still
    # works:
    wk_path = Path(wk.path)
    copy_path = wk_path.parent.joinpath(wk_path.with_suffix(".copy"))
    shutil.copytree(wk.path, copy_path, symlinks=True)
    t2_script_path_copy = Path(str(t2_script_path).replace(wk.path, f"{wk.path}.copy"))
    assert t1_script_path.read_text() == t2_script_path_copy.read_text()


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_main_script_two_actions_same_schema(null_config, tmp_path: Path):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_direct_in_direct_out.py>>",
                script_data_in="direct",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            ),
            hf.Action(
                script="<<script:main_script_test_json_in_json_out.py>>",
                script_data_in="json",
                script_data_out="json",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
                requires_dir=True,
            ),
        ],
    )
    p1_val = 101
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="main_script_test_distinct_actions_same_schema",
        path=tmp_path,
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    # check scripts generated for act 0 and 1 have different contents
    act_0_script, _ = wk.tasks.t1.template.schema.actions[0].get_script_artifact_name(
        env_spec={}, act_idx=0
    )
    act_1_script, _ = wk.tasks.t1.template.schema.actions[1].get_script_artifact_name(
        env_spec={}, act_idx=1
    )
    act_0_script_path = wk.submissions[0].scripts_path / act_0_script
    act_1_script_path = wk.submissions[0].scripts_path / act_1_script
    assert act_0_script_path.read_text() != act_1_script_path.read_text()


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_shell_env_vars(null_config, tmp_path: Path):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p1"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_shell_env_vars.py>>",
                script_data_in="direct",
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
                requires_dir=True,
            )
        ],
    )
    tasks = [
        hf.Task(
            schema=s1,
            inputs={"p1": 1},
            repeats=3,
        ),
        hf.Task(
            schema=s1,
            inputs={"p1": 1},
        ),
        hf.Task(
            schema=s1,
            inputs={"p1": 1},
            repeats=2,
        ),
    ]
    loops = [
        hf.Loop(
            tasks=[2],
            num_iterations=2,
        )
    ]
    wk = hf.Workflow.from_template_data(
        tasks=tasks,
        loops=loops,
        template_name="main_script_test_shell_env",
        path=tmp_path,
    )
    wk.add_submission(tasks=[0, 1])
    wk.submit(wait=True, add_to_known=False, status=False)  # first submission

    wk.submit(wait=True, add_to_known=False, status=False)  # outstanding runs

    for run in wk.get_all_EARs():
        run_dir = run.get_directory()
        assert run_dir
        with run_dir.joinpath("env_vars.json").open("rt") as fp:
            env_dat = json.load(fp)

        assert env_dat["HPCFLOW_WK_PATH"] == str(run.workflow.path)
        assert env_dat["HPCFLOW_WK_PATH_ARG"] == str(run.workflow.path)

        assert run.submission_idx is not None
        for js in wk.submissions[run.submission_idx].jobscripts:
            js_funcs_path = str(js.jobscript_functions_path)
            for block in js.blocks:
                for run_i in block.all_EARs:
                    if run_i.id_ == run.id_:
                        assert int(env_dat["HPCFLOW_JS_IDX"]) == js.index
                        assert env_dat["HPCFLOW_JS_FUNCS_PATH"] == js_funcs_path

        assert int(env_dat["HPCFLOW_RUN_ID"]) == run.id_
        assert int(env_dat["HPCFLOW_RUN_IDX"]) == run.index
        assert int(env_dat["HPCFLOW_RUN_PORT"]) == run.port_number

        script_name = run.get_script_artifact_name()
        sub_scripts_dir = wk.submissions[run.submission_idx].scripts_path
        script_path = sub_scripts_dir.joinpath(script_name)

        assert env_dat["HPCFLOW_SUB_SCRIPTS_DIR"] == str(sub_scripts_dir)
        assert int(env_dat["HPCFLOW_SUB_IDX"]) == run.submission_idx

        assert env_dat["HPCFLOW_RUN_SCRIPT_DIR"] == str(script_path.parent)
        assert env_dat["HPCFLOW_RUN_SCRIPT_PATH"] == str(script_path)
        assert env_dat["HPCFLOW_RUN_SCRIPT_NAME"] == script_name
        assert env_dat["HPCFLOW_RUN_SCRIPT_NAME_NO_EXT"] == script_path.stem

        assert env_dat["HPCFLOW_RUN_STD_PATH"] == str(run.get_app_std_path())
        assert (
            env_dat["HPCFLOW_RUN_LOG_PATH"]
            == env_dat["HPCFLOW_LOG_PATH"]
            == str(run.get_app_log_path())
            if run.resources.write_app_logs
            else " "
        )

        assert env_dat["HPCFLOW_ELEMENT_ID"] == str(run.element.id_)
        assert env_dat["HPCFLOW_ELEMENT_IDX"] == str(run.element.index)

        assert env_dat["HPCFLOW_ELEMENT_ITER_ID"] == str(run.element_iteration.id_)
        assert env_dat["HPCFLOW_ELEMENT_ITER_IDX"] == str(run.element_iteration.index)


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_combine_scripts_script_data_multiple_input_file_formats(
    null_config, tmp_path: Path
):
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p2"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_json_in_json_out.py>>",
                script_data_in="json",
                script_data_out="json",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
                requires_dir=True,
            ),
        ],
    )
    s2 = hf.TaskSchema(
        objective="t2",
        inputs=[
            hf.SchemaInput(parameter=hf.Parameter("p2")),
            hf.SchemaInput(parameter=hf.Parameter("p1c")),
        ],
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p3"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_hdf5_in_obj_2.py>>",
                script_data_in={"p2": "direct", "p1c": "hdf5"},
                script_data_out="direct",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
                requires_dir=True,
            ),
        ],
        parameter_class_modules=["hpcflow.sdk.core.test_utils"],
    )
    p1_val = 101
    t1 = hf.Task(schema=s1, inputs={"p1": p1_val})
    t2 = hf.Task(schema=s2, inputs={"p1c": P1(a=p1_val)})
    wk = hf.Workflow.from_template_data(
        tasks=[t1, t2],
        template_name="main_script_test",
        path=tmp_path,
        resources={"any": {"combine_scripts": True}},
    )
    wk.submit(wait=True, add_to_known=False, status=False)

    t0_p2 = wk.tasks[0].elements[0].outputs.p2
    t1_p3 = wk.tasks[1].elements[0].outputs.p3
    assert isinstance(t0_p2, hf.ElementParameter)
    assert isinstance(t1_p3, hf.ElementParameter)
    assert t0_p2.value == p1_val + 100
    assert t1_p3.value == p1_val + 100


@pytest.mark.integration
@pytest.mark.skipif("hf.run_time_info.is_frozen")
def test_combine_scripts_from_future_import(null_config, tmp_path: Path):
    s1 = hf.TaskSchema(
        objective="t1",
        actions=[
            hf.Action(
                script="<<script:import_future_script.py>>",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
            ),
        ],
    )

    wk = hf.Workflow.from_template_data(
        template_name="test_future_import",
        tasks=[hf.Task(schema=s1)],
        resources={"any": {"combine_scripts": True}},
        path=tmp_path,
    )
    wk.submit(status=False, add_to_known=False, wait=True)

    run = wk.get_EARs_from_IDs([0])[0]
    assert run.status is EARStatus.success



from __future__ import annotations
from typing import TYPE_CHECKING
import pytest
from hpcflow.app import app as hf
from hpcflow.sdk.core.errors import MalformedNestingOrderPath

if TYPE_CHECKING:
    from pathlib import Path
    from hpcflow.sdk.core.parameters import Parameter
    from hpcflow.sdk.core.types import ResourceSpecArgs
    from hpcflow.sdk.core.workflow import Workflow


@pytest.fixture
def null_config(tmp_path: Path):
    if not hf.is_config_loaded:
        hf.load_config(config_dir=tmp_path)


@pytest.fixture
def param_p1() -> Parameter:
    return hf.Parameter("p1")


@pytest.fixture
def param_p2() -> Parameter:
    return hf.Parameter("p2")


@pytest.fixture
def param_p3() -> Parameter:
    return hf.Parameter("p3")


@pytest.fixture
def workflow_w1(
    null_config, tmp_path: Path, param_p1: Parameter, param_p2: Parameter
) -> Workflow:
    s1 = hf.TaskSchema("t1", actions=[], inputs=[param_p1], outputs=[param_p2])
    s2 = hf.TaskSchema("t2", actions=[], inputs=[param_p2])

    t1 = hf.Task(
        schema=s1,
        sequences=[hf.ValueSequence("inputs.p1", values=[101, 102], nesting_order=1)],
    )
    t2 = hf.Task(schema=s2, nesting_order={"inputs.p2": 1})

    wkt = hf.WorkflowTemplate(name="w1", tasks=[t1, t2])
    return hf.Workflow.from_template(wkt, path=tmp_path)


@pytest.fixture
def workflow_w2(workflow_w1: Workflow) -> Workflow:
    """Add another element set to the second task."""
    workflow_w1.tasks.t2.add_elements(nesting_order={"inputs.p2": 1})
    return workflow_w1


def test_resources_init_equivalence_dict_list_of_obj() -> None:
    es1 = hf.ElementSet(resources={"any": {"num_cores": 1}})
    es2 = hf.ElementSet(resources=[hf.ResourceSpec(scope="any", num_cores=1)])
    assert es1 == es2


def test_resources_init_equivalence_list_list_of_obj() -> None:
    res_1_kwargs: ResourceSpecArgs = {"scope": "any", "num_cores": 1}
    es1 = hf.ElementSet(resources=[res_1_kwargs])
    es2 = hf.ElementSet(resources=[hf.ResourceSpec(**res_1_kwargs)])
    assert es1 == es2


def test_resources_init_equivalence_list_of_obj_resource_list_obj() -> None:
    res_1_kwargs: ResourceSpecArgs = {"scope": "any", "num_cores": 1}
    es1 = hf.ElementSet(resources=[hf.ResourceSpec(**res_1_kwargs)])
    es2 = hf.ElementSet(resources=hf.ResourceList([hf.ResourceSpec(**res_1_kwargs)]))
    assert es1 == es2


def test_repeats_single_int_equivalence(null_config) -> None:
    es1 = hf.ElementSet(repeats=2)
    es2 = hf.ElementSet(repeats=[{"name": "", "number": 2, "nesting_order": 0}])
    assert es1 == es2


def test_merge_envs(null_config) -> None:
    envs = {"my_env": {"version": "1.0"}}
    es = hf.ElementSet(environments=envs)
    assert es.resources.get(scope=hf.ActionScope.any()).environments == envs


def test_merge_envs_existing_any_resources(null_config) -> None:
    envs = {"my_env": {"version": "1.0"}}
    num_cores = 2
    es = hf.ElementSet(resources={"any": {"num_cores": num_cores}}, environments=envs)
    assert es.resources.get(scope=hf.ActionScope.any()).environments == envs
    assert es.resources.get(scope=hf.ActionScope.any()).num_cores == num_cores


def test_merge_envs_resource_envs_precedence(null_config) -> None:
    envs = {"my_env": {"version": "1.0"}}
    res_envs = {"other_env": {"version": "2.0"}}
    es = hf.ElementSet(resources={"any": {"environments": res_envs}}, environments=envs)
    assert es.resources.get(scope=hf.ActionScope.any()).environments == res_envs


def test_merge_envs_no_envs_with_resource_envs(null_config) -> None:
    envs = {"my_env": {"version": "1.0"}}
    es = hf.ElementSet(resources={"any": {"environments": envs}})
    assert es.resources.get(scope=hf.ActionScope.any()).environments == envs


def test_raise_env_and_envs_specified(null_config) -> None:
    with pytest.raises(ValueError):
        hf.ElementSet(env_preset="my_preset", environments={"my_env": {"version": 1}})


def test_nesting_order_paths_raise(null_config) -> None:
    with pytest.raises(MalformedNestingOrderPath):
        hf.ElementSet(nesting_order={"bad_path.p1": 1})


def test_nesting_order_paths_no_raise(null_config) -> None:
    hf.ElementSet(nesting_order={"inputs.p1": 1, "resources.any": 2, "repeats": 3})


def test_input_source_str_dict_list_str_list_dict_equivalence(null_config) -> None:
    inp_source_dict: dict[str, str | int] = {
        "source_type": "task",
        "task_source_type": "output",
        "task_ref": 0,
    }
    inp_source_str = "task.0.output"
    inp_source_list_dict = [inp_source_dict]
    inp_source_list_str = [inp_source_str]
    assert (
        hf.ElementSet.from_json_like(
            {"input_sources": {"p1": inp_source_dict}}
        ).input_sources
        == hf.ElementSet.from_json_like(
            {"input_sources": {"p1": inp_source_list_dict}}
        ).input_sources
        == hf.ElementSet.from_json_like(
            {"input_sources": {"p1": inp_source_str}}
        ).input_sources
        == hf.ElementSet.from_json_like(
            {"input_sources": {"p1": inp_source_list_str}}
        ).input_sources
    )



import os
import pytest

import hpcflow.app as hf


@pytest.mark.integration
def test_builtin_program_no_args_resource_var(new_null_config, tmp_path):
    # run a builtin program
    env_cmd = ("& " if os.name == "nt" else "") + "<<program_path>> <<args>>"
    env = hf.Environment(
        name="program_env",
        executables=[
            hf.Executable(
                label="hello_world",
                instances=[
                    hf.ExecutableInstance(
                        command=env_cmd,
                        num_cores=1,
                        parallel_mode=None,
                    )
                ],
            )
        ],
    )
    hf.envs.add_object(env, skip_duplicates=True)

    act = hf.Action(
        program="hello_world/<<resource:platform>>/hello_world<<resource:executable_extension>>",
        program_exe="hello_world",
        environments=[hf.ActionEnvironment("program_env")],
    )
    s1 = hf.TaskSchema(objective="hello", actions=[act])
    tasks = [hf.Task(s1)]
    wk = hf.Workflow.from_template_data(
        template_name="test_program_no_args",
        tasks=tasks,
        path=tmp_path,
    )
    wk.submit(wait=True, status=False)
    assert wk.submissions[0].jobscripts[0].get_stdout().strip() == "hello, world"

    hf.reload_template_components()  # remove extra env


@pytest.mark.integration
def test_builtin_program_no_args_env_var(new_null_config, tmp_path):
    # run a builtin program
    env_cmd = ("& " if os.name == "nt" else "") + "<<program_path>> <<args>>"
    default_platform = hf.ElementResources.get_default_platform()
    env = hf.Environment(
        name="program_env",
        specifiers={
            "platform": default_platform
        },  # reference this specifier in the program path
        executables=[
            hf.Executable(
                label="hello_world",
                instances=[
                    hf.ExecutableInstance(
                        command=env_cmd,
                        num_cores=1,
                        parallel_mode=None,
                    )
                ],
            )
        ],
    )
    hf.envs.add_object(env, skip_duplicates=True)

    act = hf.Action(
        program="hello_world/<<env:platform>>/hello_world<<resource:executable_extension>>",
        program_exe="hello_world",
        environments=[hf.ActionEnvironment("program_env")],
    )
    s1 = hf.TaskSchema(objective="hello", actions=[act])
    tasks = [hf.Task(s1)]
    wk = hf.Workflow.from_template_data(
        template_name="test_program_no_args",
        environments={"program_env": {"platform": default_platform}},
        tasks=tasks,
        path=tmp_path,
    )
    wk.submit(wait=True, status=False)
    assert wk.submissions[0].jobscripts[0].get_stdout().strip() == "hello, world"

    hf.reload_template_components()  # remove extra env


@pytest.mark.integration
def test_builtin_program_no_args_param_var(new_null_config, tmp_path):
    # run a builtin program
    env_cmd = ("& " if os.name == "nt" else "") + "<<program_path>> <<args>>"
    default_platform = hf.ElementResources.get_default_platform()
    env = hf.Environment(
        name="program_env",
        executables=[
            hf.Executable(
                label="hello_world",
                instances=[
                    hf.ExecutableInstance(
                        command=env_cmd,
                        num_cores=1,
                        parallel_mode=None,
                    )
                ],
            )
        ],
    )
    hf.envs.add_object(env, skip_duplicates=True)

    act = hf.Action(
        program="hello_world/<<parameter:platform>>/hello_world<<resource:executable_extension>>",
        program_exe="hello_world",
        environments=[hf.ActionEnvironment("program_env")],
    )
    s1 = hf.TaskSchema(
        objective="hello",
        actions=[act],
        inputs=[hf.SchemaInput("platform")],
    )
    tasks = [hf.Task(s1, inputs={"platform": default_platform})]
    wk = hf.Workflow.from_template_data(
        template_name="test_program_no_args",
        tasks=tasks,
        path=tmp_path,
    )
    wk.submit(wait=True, status=False)
    assert wk.submissions[0].jobscripts[0].get_stdout().strip() == "hello, world"

    hf.reload_template_components()  # remove extra env


@pytest.mark.integration
def test_builtin_program_input_output_JSON_resource_var(new_null_config, tmp_path):
    # run a builtin program that expects input and output JSON file paths as a cmdline arguments
    env_cmd = ("& " if os.name == "nt" else "") + "<<program_path>> <<args>>"
    env = hf.Environment(
        name="program_env",
        executables=[
            hf.Executable(
                label="hello_world",
                instances=[
                    hf.ExecutableInstance(
                        command=env_cmd,
                        num_cores=1,
                        parallel_mode=None,
                    )
                ],
            )
        ],
    )
    hf.envs.add_object(env, skip_duplicates=True)

    act = hf.Action(
        program="hello_world/<<resource:platform>>/hello_world<<resource:executable_extension>>",
        program_exe="hello_world",
        program_data_in="json",
        program_data_out="json",
        requires_dir=True,
        environments=[hf.ActionEnvironment("program_env")],
    )
    s1 = hf.TaskSchema(
        objective="hello",
        inputs=[hf.SchemaInput("p1"), hf.SchemaInput("p2"), hf.SchemaInput("p3")],
        outputs=[hf.SchemaInput("p4")],
        actions=[act],
    )
    p1, p2, p3 = 1, 2, 3
    tasks = [hf.Task(s1, inputs={"p1": p1, "p2": p2, "p3": p3})]
    wk = hf.Workflow.from_template_data(
        template_name="test_program_input_output_files",
        tasks=tasks,
        path=tmp_path,
    )
    wk.submit(wait=True, status=False)
    assert wk.submissions[0].jobscripts[0].get_stdout().strip() == "hello, world"
    assert wk.tasks[0].elements[0].get("outputs.p4") == p1 + p2 + p3

    hf.reload_template_components()  # remove extra env
